/*
 * cascade.c -- Functions for implementing cascading point location.
 *
 * Author -- Mike Facello
 *
 * Date -- 9/2/94
 */

#include "delcx.h"
#include "dt.h"

static int *location;		/* This array stores the location of the
				   vertices, indexed by vertex number.   */
static int *visits;		/* This array stores the number of times
                                   a vertex has been moved in the
				   cascading process.                    */

/*--------------------------------------------------------------------------*/

#define mod2       & 01

/*--------------------------------------------------------------------------*/

/* Local procedures */

static void add_vertex(int v, int e);
static int get_vertex(int e);

static void process_insert_tri(int inserted_v, int e, const int field[/*6*/]);
static void process_triangle_tri (int e, int pox, int poy, int poz);
static void process_edge_tri(int e, int yxz);
static void process_vertex_tri (int e, int new_tri);

static int redundancy_check(int v, int e);
static void print_vertices(int e);

/*--------------------------------------------------------------------------*/

int do_red = FALSE;

/*--------------------------------------------------------------------------*/

/* 
 * cascade_init () -- This function initializes the cascading method by 
 *                    placing every vertex in the infinite tetrahedron.
 */
void cascade_init (int vertex[], int n)
{
  int i;

  location = (int *) MALLOC (int, trist_max_org());
  visits = (int *) MALLOC (int, trist_max_org());
  BZERO (visits, int, trist_max_org());
  upfor (i, 1, n-4) {                      /* exclude infinite points */
    add_vertex(vertex[i], EdFacet(1, 1));
  }
}

/*--------------------------------------------------------------------------*/

void cascade_close(void)
{
  FREE(location);
  FREE(visits);
}


/*--------------------------------------------------------------------------*/
/*             PRIMITIVES FOR THE CASCADE DATA STRUCTURE                    */
/*--------------------------------------------------------------------------*/

/*
 * cascade_make () -- Creates a new triangle and initializes the stacks to
 *                    null.
 */
int cascade_make (int i, int j, int k)
{
  int ijk = trist_make (i, j, k);
  Cascade_record *cr = (Cascade_record *) trist_data_addr (TrIndex(ijk));

  cr->stack[0] = NULL;
  cr->stack[1] = NULL;

  return (ijk);
}

/*--------------------------------------------------------------------------*/

/*
 * add_vertex () -- This function pushes a vertex onto the stack
 *                  for a tetrahedron.
 */
static void add_vertex (int v, int e)
{
  Cascade_record *cr = (Cascade_record *) trist_data_addr (TrIndex(e));

#ifdef CDEBUG
  {
    int i, j, k, l;

    trist_tetra (e, &i, &j, &k, &l);

    print ("\tAdding vertex %d to tetrahedron (%d, %d, %d, %d) -- ",
	   v, i, j, k, l);
    if (    sos_positive3_inf (v,i,j,k)
	and sos_positive3_inf (v,j,i,l)
	and sos_positive3_inf (v,k,j,l)
	and sos_positive3_inf (v,i,k,l))
      print ("inside\n");
    else
      print ("OUTSIDE\n");
  }
#endif

  if (not cr->stack[e mod2])
    cr->stack[e mod2] = basic_istaque_new (1);

  basic_istaque_push (cr->stack[e mod2], v);
  if (location[v] < 0)
    location[v] = -e;
  else
    location[v] = e;
}

/*--------------------------------------------------------------------------*/

/*
 * get_vertex () -- This function pops a vertex from the tetrahedron's
 *                  stack, returning 0 if the stack is empty.
 */
static int get_vertex (int e)
{
  Cascade_record *cr = (Cascade_record *) trist_data_addr (TrIndex(e));
  Basic_istaque_adt st = cr->stack [e mod2];

  if (not st)
    return (0);
  else if (basic_istaque_empty(st)) {
    basic_istaque_dispose (st);
    cr->stack [e mod2] = NULL;
    return (0);
  }
  else
    return (basic_istaque_pop (st));
}


/*--------------------------------------------------------------------------*/
/*             SEARCH ROUTINES FOR THE CASCADE DATA STRUCTURE               */
/*--------------------------------------------------------------------------*/

/*
 * cascade_locate_vertex () -- This function locates a vertex in the
 *                             triangulation using the location array.
 *                             A negative location indicates that the
 *                             vertex is known to be redundant.  -1 is
 *                             returned for such a vertex.
 */

int cascade_locate_vertex(int v)
{
  int i, j, k, l;

  if (location[v] > 0)
    {
      trist_tetra (location[v], &i, &j, &k, &l);
      if (not (sos_in_sphere_inf_p (i,j,k,l,v))) {
#ifdef CDEBUG
	print ("&&&&& Not inserting vertex %d\n", v);
#endif
	not_inserted ++;
	vis_not_inserted += visits[v];
	add_redundant_vertex(v);
	location[v] = - location[v];
      }
    }

  if (location[v] > 0)
    return (location[v]);
  else
    return (-1);
}

/*--------------------------------------------------------------------------*/

/*
 * cascade_locate_redundant () -- This function locates a redundant vertex 
 *                                in the triangulation using the location 
 *                                array.  This only works if do_red is
 *                                true.  If a vertex is redundant, it's 
 *                                location is -location[v].
 */

int cascade_locate_redundant(int v)
{
  int i, j, k, l;

  Assert_always (location[v] < 0);

  return (- location[v]);
}

/*--------------------------------------------------------------------------*/

/*
 * cascade_update_insert () -- This function cascades vertices when an
 *                             insertion occurs.
 */
void cascade_update_insert (int v, int ijk, const int field[/*6*/], 
			    int i, int j, int k, int l)
{
  location[v] = 0;

  process_insert_tri (v, ijk, field);
  process_insert_tri (v, field[6], field);   /* jil */
  process_insert_tri (v, field[7], field);   /* kjl */
  process_insert_tri (v, field[8], field);   /* ikl */

#ifdef CDEBUG
  cascade_print_locations();
#endif
}


/*
 * process_insert_tri () -- This function will process the vertices
 *                 for a single triangle involved in a vertex insertion.
 */
static void process_insert_tri (int inserted_v, int e, const int field[/*6*/])
{
  int v;

  int ijv = field[0];
  int ikv = field[1];
  int jkv = field[2];
  int ilv = field[3];
  int jlv = field[4];
  int klv = field[5];

  int new_tri;

#ifdef CDEBUG
  print ("\nProcessing insert ");
  pr_triangle (e);
  print ("\n");

  print_vertices(e);
#endif

  while (v = get_vertex (e)) {

    if (v == inserted_v)      /* We are finished with this vertex */
      continue;

    /* Find new tetrahedron containing v */

    if (search_pos_orientation (v, jlv))
      if (search_pos_orientation (v, ilv))
	if (search_pos_orientation (v, ikv))
	  new_tri = ikv;
	else
	  new_tri = Sym (ikv);
      else 
	if (search_pos_orientation (v, ijv))
	  new_tri = ijv;
	else
	  new_tri = Sym (ijv);
    else 
      if (search_pos_orientation (v, klv))
	if (search_pos_orientation (v, jkv))
	  new_tri = jkv;
	else
	  new_tri = Sym (jkv);
      else
	if (search_pos_orientation (v, ikv))
	  new_tri = ikv;
	else
	  new_tri = Sym (ikv);

    tot_vis += 3;
    visits[v] += 3;

#ifdef REGULAR
    if (redundancy_check (v, new_tri))
      if (not do_red)                     /* Continue the point location for
			 		     this vertex, even though it is
					     redundant */
	continue;
#endif

    add_vertex (v, new_tri);
  }

}




/*--------------------------------------------------------------------------*/

/*
 * cascade_update_triangle () -- This function cascades vertices when a
 *                               triangle-to-edge flip  occurs.
 */
#ifdef OLDSTUFF
void cascade_update_triangle (int xyz, const int field[/*6*/], 
			      int x, int y, int z, int o, int p)
#endif
void cascade_update_triangle (int x, int y, int z, int o, int p,
			      const int field[/*10*/])
{
  int j;

#ifdef OLDSTUFF
  int pox = Enext (Turn (field[2]));  /* field[2] = zxo */
  int poy = Fnext(pox);
  int poz = Fnext(poy);
#endif

  int pox = Sym(Enext2(field[1]));
  int poy = Sym(Enext2(field[2]));
  int poz = Sym(Enext2(field[3]));

  Assert (dt_test_triangle (pox, p, o, x));
  Assert (dt_test_triangle (poy, p, o, y));
  Assert (dt_test_triangle (poz, p, o, z));

  process_triangle_tri (field[0] /*xyz*/, pox, poy, poz);
  process_triangle_tri (Sym (field[0]) /*Sym(xyz)*/, pox, poy, poz);

  /* The field array contains the 6 outward pointing triangles of the double
     tetrahedron for this flip.                                             */

#ifdef OLDSTUFF
  upfor (j, 0, 5)
    process_triangle_tri (Sym(field[j]), pox, poy, poz);
#endif
  upfor (j, 4, 9)
    process_triangle_tri (field[j], pox, poy, poz);

  trist_delete (field[0]);

#ifdef CDEBUG
  cascade_print_locations();
#endif
}


/*
 * process_triangle_tri () -- This function will process the vertices
 *                 for a single triangle involved in a T-to-E flip.
 */
static void process_triangle_tri (int e, int pox, int poy, int poz)
{

  int v;
  int new_tri;

#ifdef CDEBUG
  print ("\nProcessing triangle ");
  pr_triangle (e);
  print ("\n");

  print_vertices(e);
#endif

  while (v = get_vertex (e)) {

    /* Find new tetrahedron containing v */

    if (search_pos_orientation (v, pox))
      if (search_pos_orientation (v, poy))
	new_tri = poy;
      else
	new_tri = Sym (poy);
    else 
      if (search_pos_orientation (v, poz))
	new_tri = poz;
      else
	new_tri = Sym (poz);

    tot_vis += 2;
    visits[v] += 2;

#ifdef REGULAR
    if (redundancy_check (v, new_tri))
      if (not do_red)                    /* Continue the point location for
					    this vertex, even though it is
					    redundant */
	continue;
#endif

    add_vertex (v, new_tri);
  }

}




/*--------------------------------------------------------------------------*/

/*
 * cascade_update_edge () -- This function cascades vertices when an
 *                           edge-to-triangle flip  occurs.
 */
#ifdef OLDSTUFF
void cascade_update_edge (int pox, int poy, int poz, const int field[/*6*/], 
			  int x, int y, int z, int o, int p)
#endif

void cascade_update_edge (int x, int y, int z, int o, int p,
			  const int field[/*10*/])
{
  int j;
#ifdef OLDSTUFF
  int yxz = Fnext (Sym (field[0]));   /* field[0] = xyo */
#endif
  int yxz = Sym (field[3]);           /* field[3] = xyz */
  Assert (dt_test_triangle (yxz, y, x, z));

#ifdef OLDSTUFF
  process_edge_tri (pox, yxz);
  process_edge_tri (Sym (pox), yxz);
  process_edge_tri (poy, yxz);
  process_edge_tri (Sym (poy), yxz);
  process_edge_tri (poz, yxz);
  process_edge_tri (Sym (poz), yxz);
#endif

  process_edge_tri (field[0], yxz);
  process_edge_tri (Sym (field[0]), yxz);
  process_edge_tri (field[1], yxz);
  process_edge_tri (Sym (field[1]), yxz);
  process_edge_tri (field[2], yxz);
  process_edge_tri (Sym (field[2]), yxz);

#ifdef OLDSTUFF
  upfor (j, 0, 5)
    process_edge_tri (Sym(field[j]), yxz);
#endif

  upfor (j, 4, 9)
    process_edge_tri (field[j], yxz);

  trist_delete(field[0]);   /* pox */
  trist_delete(field[1]);   /* poy */
  trist_delete(field[2]);   /* poz */

#ifdef CDEBUG
  cascade_print_locations();
#endif
}


/*
 * process_edge_tri () -- This function will process the vertices
 *                 for a single triangle involved in a T-to-E flip.
 */
static void process_edge_tri (int e, int yxz)
{

  int v;
  int new_tri;

#ifdef CDEBUG
  print ("\nProcessing edge ");
  pr_triangle (e);
  print ("\n");

  print_vertices(e);
#endif

  while (v = get_vertex (e)) {

    /* Find new tetrahedron containing v */

    if (search_pos_orientation (v, yxz))
      new_tri = yxz;
    else
      new_tri = Sym (yxz);

    tot_vis += 1;
    visits[v] += 1;

#ifdef REGULAR
    if (redundancy_check (v, new_tri)) {
      if (not do_red)                    /* Continue the point location for
					    this vertex, even though it is
					    redundant */
	continue;
    }
#endif


    add_vertex (v, new_tri);
  }

}




/*--------------------------------------------------------------------------*/

/*
 * cascade_update_vertex () -- This function cascades vertices when an
 *                             vertex flip  occurs.
 */
void cascade_update_vertex (const int field[/*9*/])
{
  int i, v;
  int new_tri = field[6];

  v = Dest(field[0]);

  upfor (i, 0, 5) {
    process_edge_tri (field[i], new_tri);
    process_edge_tri (Sym(field[i]), new_tri);
    trist_delete (field[i]);
  }

  location[v] = -new_tri;
  add_vertex(v, new_tri);

#ifdef CDEBUG
  cascade_print_locations();
#endif
}


/*
 * process_vertex_tri () -- This function will process the vertices
 *                 for a single triangle involved in a vertex flip.
 */
static void process_vertex_tri (int e, int new_tri)
{

  int v;

#ifdef CDEBUG
  print ("\nProcessing vertex ");
  pr_triangle (e);
  print ("\n");

  print_vertices(e);
#endif

  while (v = get_vertex (e)) {

    /* Every vertex is immediatly put into new_tri */

#ifdef REGULAR
    if (redundancy_check (v, new_tri))
      if (not do_red)                    /* Continue the point location for
					    this vertex, even though it is
					    redundant */
	continue;
#endif

    add_vertex (v, new_tri);
  }

}


/*--------------------------------------------------------------------------*/

/*
 * redundancy_check () -- This function checks if a vertex is redundant
 *                        based on the current tetrahedron that contains
 *                        it.  Note that this procedure is only used
 *                        during the cascading part of the search, not
 *                        for the final redundancy check before the point
 *                        is inserted.
 */

static int redundancy_check (int v, int e)
{
  int i, j, k, l;

/*  if (((visits[v] - 3) div redcheck_dist) < (visits[v] div redcheck_dist)) */
/*  To do above, I'd need to add increment as a parameter.                   */

  if (location[v] < 0)
    return (1);

  if ((visits[v] % redcheck_dist) == 0) {

    trist_tetra (e, &i, &j, &k, &l);
    in_tests ++;

#ifdef SDEBUG
    print ("Testing that point is inside tetrahedron (%d, %d, %d, %d)\n",
	   i, j, k, l);
    Assert_always (    sos_positive3_inf (v,i,j,k)
		   and sos_positive3_inf (v,j,i,l)
		   and sos_positive3_inf (v,k,j,l)
		   and sos_positive3_inf (v,i,k,l));
#endif
    if (not sos_in_sphere_inf_p(i, j, k, l, v)) {
#ifdef SDEBUG
      print ("&&&&& Not inserting vertex %d\n", v);
#endif
      add_redundant_vertex(v);

      vis_not_inserted += visits[v];
      not_inserted++;   /* This point is "flipped out" of the triang.  */

      location[v] = - location[v];

      return (1);
    }
  }

  return (0);
}

/*--------------------------------------------------------------------------*/

#ifdef CDEBUG

/*
 * print_vertices () -- This function will print the vertices in the stack
 *                      for a triangle.
 */
static void print_vertices (int e)
{
  Cascade_record *cr = (Cascade_record *) trist_data_addr (TrIndex(e));
  Basic_istaque_adt st = cr->stack [e mod2];
  int i, j, k;

  trist_triangle (e, &i, &j, &k);
  print ("Printing stack for (%d, %d, %d):\n", i, j, k);

  if (not st)
    print ("NO STACK ALLOCATED");
  else
    basic_istaque_print (st, stdout);

  print ("\n\n");
}

/*--------------------------------------------------------------------------*/

/*
 * cascade_print_locations () --
 */

void cascade_print_locations (void)
{
  int v, i, j, k, l;

  print ("\n\nLocations of vertices:\n\n");

  upfor (v, 1, trist_max_org() - 4)
    {
      if (location[v] > 0)
	{
	  trist_tetra (location[v], &i, &j, &k, &l);
	  print ("Vertex %d: (%d, %d, %d, %d) -- ", v, i, j, k, l);
	  if (    sos_positive3_inf (v,i,j,k)
	      and sos_positive3_inf (v,j,i,l)
	      and sos_positive3_inf (v,k,j,l)
	      and sos_positive3_inf (v,i,k,l))
	    print ("inside, ");
	  else
	    print ("OUTSIDE, ");
	  print ("(i,j,k,l) = %d\n", sos_positive3_inf (i,j,k,l));
	}
      else if (location[v] == 0)
	print ("Vertex %d:  ALREADY INSERTED\n", v);
      else
	{
	  trist_tetra (-location[v], &i, &j, &k, &l);
	  print ("Vertex %d: (%d, %d, %d, %d) -- ", v, i, j, k, l);
	  if (    sos_positive3_inf (v,i,j,k)
	      and sos_positive3_inf (v,j,i,l)
	      and sos_positive3_inf (v,k,j,l)
	      and sos_positive3_inf (v,i,k,l))
	    print ("inside ");
	  else
	    print ("OUTSIDE ");
	  print ("and REDUNDANT\n", v);
	}
    }
}

#endif
