/* delcx/delcx.c  ---  3-d Delaunay Simplicial Complex (random incremental) */
/* revised by Mike Facello from detri.c by Ernst Mucke */

/*--------------------------------------------------------------------------*/

#include "delcx.h"

static char version[] = "@(#) Delcx 4.1";
static char purpose[] = "3D Delaunay Simplicial Complex";
static char algorithm[] = "Randomized Incremental Flip Algorithm";
static char author[] = "Michael Facello";
static char alf__copyright[] = "";

/*--------------------------------------------------------------------------*/

/* program parameters */

static char usage[] = "\n\
usage: \n\
       %s [OPTIONS] [-s <seed>] <DATA> \n\
with:  \n\
       <DATA> ........ path name of input data \n\
OPTIONS: \n\
       -s <seed> ..... seed for random number generator \n\
       -t ............ test option \n\
       -r ............ force use of regular triangulation algorithm \n\
       -n ............ no randomization \n\
       -b ............ no binary output \n\
       -k ............ keep SoS artefacts \n\
       -d ............ use history dag for search \n\
       -X <option> ... undocumented options \n\
ABC mode:             (All But Computation) batch mode: \n\
       -A ............. read commands from standard input, \n\
       -B <FILE> ...... from given <FILE>, or \n\
       -C <COMMAND> ... from command-line. \n\
";

/*--------------------------------------------------------------------------*/

/* program flags/parameters and their initial/default setting */

static char *cmd_name, *data_name, *abc_file_name;

       int do_test = FALSE;
       int do_output_red = FALSE;
       int do_randomize = TRUE;
static int do_binary = TRUE;
static int do_peel = TRUE;
  
int Seed;

/* Made global for ROMAN */
Dt_input_scan *data;

static int maxtri_code = 1;  	/* see maxtri() */
/* note: By default maxtri_code==1, and maxtri (n) returns some (empirical)
   number suitable for the iniltial Trist malloc.  If maxtri_code==0, the
   correct upper bound according to the Upper Bound Theorem is used.  This,
   however, is no longer necessary, because Trist is now semi-dynamic.
   Cf: maxtri(), -X ubt option. */


/*--------------------------------------------------------------------------*/

/* local procedures (w/ common variables) */

static int  maxtri(int n);
static void delcx(void);
static void closest(int n, int max_n, int vertex[], Trist_num *q, 
		     int hull_vertex[]);
static Trist_num aftermath (int *hull_ef, Trist_num num1);

static void print_quantities(Trist_num tq, Infinite_num iq);
static void open_info_file(char data_title[]);
static void print_info_file_dt(char diagram[], Trist *s, Trist_num q1, 
				Trist_num q2, Infinite_num iq, 
				Search_info si, double time);
static void print_info_file_ch(Trist_num q1, Trist_num q2);
static void close_info_file(double total_time, double terminal_time, 
			     Trist_num q);

#ifdef __DEBUG__
static void dummy_flip_hook(char code, int x, int y, int z, int o, int p, 
			     int ref);
#endif

static void parse_abc_option(char *current, char option, char string[]);
static void  abc_mode(void);


static FILE *info_file;
static int info_trist_bpt = 0, info_trist_bytes = 0;

static int abort_flag = FALSE;  /* see delcx() and delcx_error() */

/*--------------------------------------------------------------------------*/

/*
 * main -- This is the routine that is first called in the delcx program.
 *
 */
main (int argc, char *argv[])
{
  int c;
  int is_wrong = FALSE;
  char abc_option = ' ';
  print ("%s -- %s (%s)\n%s\n%s\n",
         basic_strip (version), purpose, algorithm, author,alf__copyright);
  basic_error_hook (delcx_error);
  regular_trace_mode = 1;
  ;
#ifdef __DEBUG__
/*  basic_malloc_debug (1); */

  {
    /* medium */
    regular_test_flag = TRUE;
    flip_test_flag = TRUE;    
    trist_test_flag = TRUE;   
    trist_io_check_flag = TRUE;
    ;
    /* extreme*/
    regular_proto_flag = FALSE;
    flip_proto_flag = FALSE;
    dt_test_proto_flag = FALSE;
    sos_test_flag = FALSE;
    ;
  }
#endif
  ;
  cmd_name = argv[0];
  (void) basic_getarg_init (argc - 1, argv + 1);

  {
    /***** Seed Initialization *****/
    Seed = basic_seed();
  }

  srand(Seed);
  while ((c = basic_getarg ("trnbvkdos:X:AB:C:")) != 0)
    switch (c)
      {
      case 't':   /* -t ... call dt_test_pedantic() */
        do_test = TRUE;
	do_red = TRUE;
        break;
      case 'r':	/* -r ... force use of regular triangulation algorithm */
	regular_triang = TRUE;
	break;
      case 'n':	/* -n ... no randomization */
	do_randomize = FALSE;
	break;
      case 'b':	/* -b ... no binary */
	do_binary = FALSE;
      case 'k': /* -k ... don't remove SoS artifacts */
	do_peel = FALSE;
	break;
      case 'd': /* -d ... use tridag for search */
	search_type = TRIDAG_SEARCH;
	break;
      case 'o': /* -o ... output redundant vertices */
	do_output_red = TRUE;
	do_red = TRUE;
	break;
      case 's':   /* -s <seed> ... initializes the seed */
	if (sscanf (basic_getarg_optarg, "%d", &Seed) != 1)
	  basic_error ("unrecognized option: -s %s", basic_getarg_optarg);
	srand(Seed);
	break;

/*      case 'u':
	if (sscanf (basic_getarg_optarg, "%d", &upper_bound_multiplier) != 1)
	  basic_error ("unrecognized option: -u %s", basic_getarg_optarg);
	break; */

      case 'X':
        { /* -X <OPTION> ... undocumented options */
          if (strcmp (basic_getarg_optarg, "upt") == 0)
            { /* -X upt ... use UPT for initial Trist malloc */
              maxtri_code = 0;
              (void) maxtri (1); /* just for testing */
            }
          else
            is_wrong = TRUE;
          break;
        }

      case -1:    /* <data_name> */
        data_name = basic_getarg_optarg;
        break;
      case 'A':
      case 'B':
      case 'C':
        /* -A  or
           -B <COMMAND_FILE>  or
           -C <COMMAND_STRING> ... read binaries and execute commands */
        parse_abc_option (&abc_option, c, basic_getarg_optarg);
        break;
      case '?':
      default:
        is_wrong = TRUE;
      }
  if (is_wrong or (not data_name))
    basic_error (usage, cmd_name);
  if (not (basic_access (data_name)))
    basic_error ("Can't find %s.", data_name);
  abort_flag = TRUE; /* from now on we core-dump on errors */
  ;
#ifdef __SOS_TRACE__
  /* just in case we want to trace the SoS routines... */
  sos_trace_file = basic_fopen (basic_cb_frmt ("%s.sos", data_name), "w");
#endif
  ;
  if (abc_option == ' ')
    delcx ();
  else
    abc_mode ();

  /* shutdown */

  basic_malloc_info_print (stdout);
  if (abc_option == 'C')
    basic_rm (abc_file_name);

  return (0);
}

/*--------------------------------------------------------------------------*/

/*
 * maxtri (n) -- returns (expected) maximum number of triangles for n 
 *               vertices; global maxtri_code selects between true and 
 *               "empirical" upper bounds 
 */
static int maxtri (int n)
{
  int u;
  switch (maxtri_code)
    {
     case 0:  /* true upper bound */
      u = trist_upper_bound (n);
      break;
     case 1: /* linear "empirical" bound */
      u = 1 * n;      /*  used to be 15 * n  */
      break;
     default:

      basic_error ("unknown maxtri_code: %d", maxtri_code);
    }
  return (u);
}

/*--------------------------------------------------------------------------*/

/*
 * delcx -- The main routine for the Del. Simplicial Complex.  This is the
 *          routine called for the non-graphical version of delcx.
 */
static void delcx (void)
{
  double time_total = basic_utime ();
  double time_terminal = basic_seconds ();
  Trist_num q;
  int *vertex, *hull_vertex;
  int i, n;
  /*  Dt_input_scan *data; */

  /* First scan and then read in the data */
  
  data = dt_input_scan (data_name);

  if (data->n < 4) {
    print ("\nInput error: point set must contain at least 4 vertices\n");
    exit(0);
  }

  sos_matrix (data->n + 4, 4, data->scale,
            Lia_DIGITS (5 * data->decimals + 3),   /* sos_in_sphere() */
            Lia_DIGITS (2 * data->decimals + 1));  /* sum of squares */
  dt_input_load (data_name, data);
  if (data->has_weights)
    regular_triang = TRUE;
  ;

  /* Initialize some variables */
  
  open_info_file (data->title);
  vertex = MALLOC (int, 1 + data->n);
  hull_vertex = MALLOC (int, 1 + data->n);
  open_flip (data->n);
  open_infinite (data->n);
  basic_cb_clear ();
  open_info_file (basic_cb_str ());

  /* Miscellaneous other stuff here in delcx -- remove duplicate
     vertices and generate a random ordering for inserting the vertices. */

  scan_vertices (data->n, &n, vertex, 1, FALSE);
  n += 4;			                 /* Add in infinite points */
  print("Number of vertices: %d\n", n);
  printf("The seed is %d\n", Seed);
  randomize (vertex, n-4);

#ifdef SDEBUG
  {
    int i, j;

    upfor (i, 1, n)
      {
	print ("%d: ", i);
	upfor (j, 1, 3)
	  print ("%s ", lia_deci (sos_lia (i, j)));
	print ("\n");
	lia_clear ();
      }
  }
#endif

  /* Compute the triangulation */
  
  closest (n, data->n+4, vertex, &q, hull_vertex);

  /* Close everything down and print out statistics */

  FREE (hull_vertex);
  FREE (vertex);

  close_flip ();
  sos_shutdown ();
  
  time_total = basic_utime () - time_total;
  time_terminal = basic_seconds () - time_terminal;
  close_info_file (time_total, time_terminal, q);
}

/*--------------------------------------------------------------------------*/

/*
 * closest (n, max_n, vertex, q, hull_vertex) -- This procedure computes
 *              the triangulation and computes statistics.  Note that most
 *              of the work is done in the compute_triangulation()
 *              procedure.
 */
static void closest (int n, int max_n, int vertex[], Trist_num *q, 
		     int hull_vertex[])
     /* output -- q, hull_vertex */
{
  char *diagram;
  char *path_name = STRDUP (dt_PATH (data_name));
  Trist *s;
  Trist_num num1, num2;
  double utime;
  Infinite_num iq;
  Search_info si;

  int hull_ef;
  
#ifdef REGULAR
  print ("Regular triangulation: %d vertices, <= %d triangles ...\n",
	 n, trist_upper_bound (n));
  diagram = "Regular triangulation";
#else
  print ("Delaunay triangulation: %d vertices, <= %d triangles ...\n", 
	 n, trist_upper_bound (n));
  diagram = "Delaunay triangulation";
#endif

  /* Allocate the trist structure */

  trist_set (s = trist_alloc (max_n, maxtri (n)));
  (void) dt_test_hooks (sos_positive3_inf, sos_in_sphere_inf);

  /* Perform the triangulation */

  utime = basic_utime ();
  compute_triangulation (vertex, n);
  utime = basic_utime () - utime;

  /* Get statistics on history dag and remove it */

  if (do_red)
    compute_redundant_loc();

  if (do_test) {
    (void) check_redundancies ();
  }
  si = search_info (n);
  search_close ();

  num1 = trist_num ();
  iq = infinite_num (num1);

  if (do_test) {		/* test with infinite triangles */
    (void) dt_test_pedantic (vertex, n-4, TRUE);
  }
  
  /* Clean up the triangulation by removing all infinite triangles. */

  kill_infinite_triangles ();

  /* More statistics ... */

  num1 = trist_num ();
  print_quantities (num1, iq);
  if (do_test) {		/* test without infinite triangles */
    (void) dt_test_hooks (sos_positive3, sos_in_sphere);
    (void) dt_test_pedantic (vertex, n-4, TRUE);
  }

  num2 = aftermath (&hull_ef, num1);
  *q = num2;
  
  trist_data_size (0);
  
  if (do_binary) {
    trist_current()->max_org -= 4;	     /* Exclude infinite points */
    if (regular_triang and do_output_red)
      trist_permute_hook (update_redundant_loc);

    /* Save the triangulation */
    dt_save (path_name, If(data->has_weights,DT_WEIGHTED,DT_CLOSEST), s, q);

    /* Save the locations of redundant points */
    if (regular_triang and do_output_red)
      output_redundant_loc (data_name);
  }

  print_info_file_dt (diagram, s, num1, num2, iq, si, utime);
  print_info_file_ch (num1, num2);

  /* Deallocate everything */

  trist_kill (s);
}

/*--------------------------------------------------------------------------*/

static Trist_num aftermath (int *hull_ef, Trist_num num1)
     /* Refinement of closest/furthest_point(). */
{
  print ("Done (f=%d, t=%d, %d or %0.2f%% degenerated).\n",
         num1.f, num1.t, num1.t_flat, 100.0 * num1.t_flat / num1.t);
  if (do_peel and (num1.t_flat > 0) and (num1.t_proper > 0))
    {
      *hull_ef = trist_peel (-1, num1.t_flat, num1.fh);
      if (do_test)
        (void) dt_test ();
      return (trist_num ());
    }
  return (num1);
}

/*--------------------------------------------------------------------------*/

static void print_quantities (Trist_num tq, Infinite_num iq)
{
  print ("--  quantities before peeling: \n");
  print ("         Inner : v = %d, e = %d, f = %d, t = %d\n", tq.v, tq.e, tq.f,
	 tq.t);
  print ("         Outer : v = 4, e = %d, f = %d, t = %d\n", iq.ext_e,
	 iq.ext_f, iq.ext_t);
  print ("         Hull  : v = %d, e = %d, f = %d\n", tq.vh, tq.eh,
	 tq.fh);
  ;
}


/*--------------------------------------------------------------------------*/

static void open_info_file (char data_title[])
{
  char *path = STRDUP (info_PATH (data_name));
  info_file = basic_fopen (path, "w");
  fprint (info_file, "#\n# %s (%s), on %s, %s\n#\n  Data: %s\n  Title: %s\n",
          basic_strip (version), cmd_name, basic_hostname (), basic_date (),
          data_name, data_title);
}

/*--------------------------------------------------------------------------*/

static void print_info_file_dt (char diagram[], Trist *s, Trist_num q1, 
				Trist_num q2, Infinite_num iq, 
				Search_info si, double time)
{
  Trist_info *ti = trist_info ();
  Flip_info *fi = flip_info ();

  fprint (info_file, "\n\n\n\n* Delaunay Simplicial Complex\n");
  fprint (info_file, "%12d . vertices\n", q2.v);
  fprint (info_file, "%12.3f . cpu secs\n", time);
  fprint (info_file, "%12d . transform mainloops\n", fi->mainloops);
  fprint (info_file, "%12d . triangle-to-edge flips\n", fi->tflips);
  fprint (info_file, "%12d . edge-to-triangle flips\n", fi->eflips);

#ifdef REGULAR
  fprint (info_file, "%12d . vertex flips\n", fi->vflips);
#endif

  fprint (info_file, "%12d . skips\n", fi->skips);
  if (do_peel)
    {
      fprint (info_file, "%12d . tetrahedra (was: %d, flat: %d)\n",
              q2.t, q1.t, q1.t_flat);
      fprint (info_file, "%12d . triangles (was: %d, last: %d, max: %d)\n",
              q2.f, q1.f, s->last_triangle, s->max_triangle);
      fprint (info_file, "%12d . edges (was: %d)\n", q2.e, q1.e);
    }
  else
    {
      fprint (info_file, "%12d . tetrahedra (flat: %d)\n",
              q2.t, q1.t_flat);
      fprint (info_file, "%12d . triangles (last: %d, max: %d)\n",
              q2.f, s->last_triangle, s->max_triangle);
      fprint (info_file, "%12d . edges \n", q2.e);
    }
  fprint (info_file, "%12s . Orgs\n",     basic_counter (ti->orgs));
  fprint (info_file, "%12s . Syms\n",     basic_counter (ti->syms));
  fprint (info_file, "%12s . Enexts\n",   basic_counter (ti->enexts));
  fprint (info_file, "%12s . Fnexts\n",   basic_counter (ti->fnexts));
  fprint (info_file, "%12s . Fsplices\n", basic_counter (ti->fsplices));
  info_trist_bpt = ti->bpt;
  
  /* Added for Random Incremental dt */

  fprint (info_file, "* External facets\n");
  fprint (info_file, "%12d . tetrahedra\n", iq.ext_t);
  fprint (info_file, "%12d . triangles\n",  iq.ext_f);
  fprint (info_file, "%12d . edges\n",      iq.ext_e);

  if (search_type == TRIDAG_SEARCH) {
    fprint (info_file, "* Search statistics\n");
    fprint (info_file, "%12.3f . Average search length\n", si.avg_depth);
    fprint (info_file, "%12d . Height of dag\n", si.height);
    fprint (info_file, "%12d . Triangles in dag\n", si.hier_size);
  }

#ifdef REGULAR
  fprint (info_file, "%12d . Vertices not inserted\n", si.not_inserted);
  fprint (info_file, "%12.3f . Average search length when not inserted\n",
	  si.avg_depth_ninsert);
  fprint (info_file, "%12d . In-sphere tests\n", 
	  si.in_tests);
#endif

  info_trist_bytes += trist_last() * ti->bytes;
}

/*--------------------------------------------------------------------------*/

static void print_info_file_ch (Trist_num q1, Trist_num q2)
{
  fprint (info_file, "* Convex hull\n");
  if (do_peel)
    {
      fprint (info_file, "%12d . vertices (was: %d)\n", q2.vh, q1.vh);
      fprint (info_file, "%12d . edges (was: %d)\n", q2.eh, q1.eh);
      fprint (info_file, "%12d . triangles (was: %d)\n", q2.fh, q1.fh);
    }
  else
    {
      fprint (info_file, "%12d . vertices\n", q2.vh);
      fprint (info_file, "%12d . edges\n", q2.eh);
      fprint (info_file, "%12d . triangles\n", q2.fh);
    }
}

/*--------------------------------------------------------------------------*/

static void close_info_file (double total_time, double terminal_time, 
			     Trist_num q)
{
  Lia_info *li = lia_info();
  Basic_malloc_info *mi = basic_malloc_info ();
  int i;
  double util0_c;

  util0_c = (100.0 * q.f) / (double) trist_upper_bound (q.v);

  fprint (info_file, "* Lia counters\n");
  fprint (info_file, "%12s . Lia mul calls\n", basic_counter (li->mul_calls));
  fprint (info_file, "%12s . Lia mul elops\n", basic_counter (li->mul_elops));
  fprint (info_file, "%12s . Lia add calls\n", basic_counter (li->padd_calls));
  fprint (info_file, "%12s . Lia add elops\n", basic_counter (li->padd_elops));
  fprint (info_file, "%12s . Lia sub calls\n", basic_counter (li->psub_calls));
  fprint (info_file, "%12s . Lia sub elops\n", basic_counter (li->psub_elops));
  fprint (info_file, "%12d . Lia maximum digit\n", li->maximum_digit);
  sos_depth_counters_output (info_file);
  infinite_summary (info_file);

  fprint (info_file, "* Flip vs. Search\n\n");
  fprint (info_file, "    Positive3 tests\n");
  fprint (info_file, "%54s\n", "Number of Infinite Points");
  fprint (info_file, "\t\t\t%4d %8d %8d %8d\n", 0, 1, 2, 3);
  fprint (info_file, "\tSearch\t\t");
  for (i = 0; i < 4; i++)
    fprint (info_file, "%8d ", sos_pos3 [i][SEARCH]);
  fprint (info_file, "\n");
  fprint (info_file, "\tSucc. flip\t");
  for (i = 0; i < 4; i++)
    fprint (info_file, "%8d ", sos_pos3 [i][SUCC_FLIP]);
  fprint (info_file, "\n");
  fprint (info_file, "\tUnsucc. flip\t");
  for (i = 0; i < 4; i++)
    fprint (info_file, "%8d ", sos_pos3 [i][UNSUCC_FLIP]);
  fprint (info_file, "\n\n");

  fprint (info_file, "* Miscellaneous\n");
  fprint (info_file, "%12.3f . %% UBT utilization\n",
          util0_c);
  fprint (info_file, "%12d . bytes per triangle\n",
          (int) sizeof (Trist_record));
  fprint (info_file, "%12.3f . Mb SoS\n", basic_mbytes (sos_bytes ()));

  if (search_type == TRIDAG_SEARCH)
    fprint (info_file, "%12.3f . Mb Tridag (%d bytes per triangle)\n",
	    basic_mbytes (info_trist_bytes), info_trist_bpt);
  else
    {
      fprint(info_file, "%12.3f . Mb Trist + Cascade record (%d bytes per triangle)\n",
	     basic_mbytes (info_trist_bytes), info_trist_bpt);
    }

  fprint (info_file,
          "%12.3f . Mb malloc, total (arena: %.3f Mb)\n",
          basic_mbytes (mi->total), basic_mbytes (mi->arena));
  dt_print_info_sec (info_file, total_time,    " CPU secs, delcx total");
  dt_print_info_sec (info_file, terminal_time, "real secs, delcx total");

  print ("Needed %0.3f sec\n\n",
         terminal_time);

  basic_fclose (info_file);
}

/*--------------------------------------------------------------------------*/

void delcx_error (const char message[])
{
  fprint (stderr, "\n%s: %s\n", cmd_name, message);
  if (info_file)
    {
      fprint (info_file, "\nERROR.\n%s\n", message);
      basic_fclose (info_file);
    }
  if (abort_flag)
    abort ();
  else
    exit (1);
}

/*--------------------------------------------------------------------------*/

#ifdef __DEBUG__

static void dummy_flip_hook (char code, int x, int y, int z, int o, int p, 
			     int ref)
{
  switch (code)
    {
     case FLIP_TRIANGLE:
      (void) dt_test_triangle (ref, p, o, x);
      break;
     case FLIP_EDGE:
      (void) dt_test_triangle (ref, x, y, z);
      break;
     default:
      Assert (FALSE);
    }
}

#endif

/*--------------------------------------------------------------------------*/

static void parse_abc_option (char *current, char option, char string[])
     /* Refinement of main(). */
{
  static char internal[30];
  FILE *f;
  if ((*current != ' ') and (*current != option))
    basic_error ("Can't use -A, -B, and -C options together: -%c %s",
                 option, string);
  if ((*current == ' ') and (option == 'C'))
    { /* start new internal command file */
      sprintf (internal, ".detri.%d.abc", basic_seed ());
      abc_file_name = internal;
      f = basic_fopen (abc_file_name, "w");
      fprint (f, "%s\n", string);
      basic_fclose (f);
    }
  else if (option == 'C')
    { /* continue internal command file */
      f = basic_fopen (abc_file_name, "a");
      fprint (f, "%s\n", string);
      basic_fclose (f);
    }
  else if (option == 'B')
    { /* take given command file */
      abc_file_name = string;      
    }
  else
    { /* read from standard input */
      abc_file_name = ".";
    }
  *current = option;
}

/*--------------------------------------------------------------------------*/

static void abc_mode (void)
     /* Detri [ABC] batch mode. (All but computation. :) */
{
  int i;
  Dt *dt;
  char *buf, *token[99];
  char *dt_path = STRDUP (dt_PATH (data_name));
  Dt_input_scan *data;
  FILE *f = If ((strcmp (abc_file_name, ".") == 0),
                stdin, basic_fopen (abc_file_name, "r"));

  data = dt_input_scan (data_name);
  sos_matrix (data->n, 4, data->scale,
            Lia_DIGITS (5 * data->decimals + 3),   /* sos_in_sphere() */
            Lia_DIGITS (2 * data->decimals + 1));  /* sum of squares */
  dt_input_load (data_name, data);
  dt = dt_load (dt_path);
  print ("Entering [ABC] batch mode...\n", basic_strip (version));
  while (buf = basic_cb_getline (f))
    {
      buf[strlen (buf) - 1] = '\0';  /* delete '\n' from basic_cb_getline() */
      i = basic_tokenize (buf, token, 99);
      if (basic_arg_find ("print", i, token))
        dt_print_cmd (data_name, dt, i, token);
      else if (basic_arg_find ("help", i, token))
        dt_print_cmd (data_name, dt, i, token);
    }
  if (f and (f != stdin))
    basic_fclose (f);
  dt_kill (dt);
  sos_shutdown ();
}

/*--------------------------------------------------------------------------*/
