/* delcx/delcx.h --- main header file */

#ifndef __DELCX_H__  /* Include only once! */
#define __DELCX_H__  1

#include "basic.h"
#include "dt.h"
#include "infinite.h"

/*------------------------------------------------------------------------*/
/*                              delcx.c                                  */

extern int regular_triang;		/* using regular triangulation or
					   Delaunay triangulation?        */
extern int search_type;
extern int do_test;
extern int do_red;
extern int do_output_red;

void delcx_error(const char message[]);

/*------------------------------------------------------------------------*/
/*                             regular.c                                  */

void compute_triangulation (int vertex[], int n);

extern int regular_test_flag, regular_proto_flag, regular_trace_mode;

/*------------------------------------------------------------------------*/
/*                              flip.c                                    */

typedef struct flip_info_type
{
  int skips, eflips, tflips, vflips, mainloops;
} Flip_info;


typedef struct redundant_info_type
{
  int vflips;
  Trist_org *flipped_vert;
  int *flipped_loc;
} Redundant_info;
    
#define FLIP_INSERT   'I'
#define FLIP_TRIANGLE 'T'
#define FLIP_EDGE     'E'
#define FLIP_VERTEX   'V'
#define FLIP_RESET_INFO  'R'
#define FLIP_RETURN_INFO ' '

void open_flip(int n), close_flip(void);
void construct_outer_tetra(void);
int flip_transform(void);

void delete_the_triangle(int e);
void set_blue(int t);		/* used in search.c */
int is_regular(int xyz);
void flip_insert(int ijk, int v);

/* Functions for handling redundant vertices */

void add_redundant_vertex(Trist_org v);
void compute_redundant_loc(void);
int update_redundant_loc(const int map[]);
void output_redundant_loc(char data_name[]);
Redundant_info *input_redundant_loc(char data_name[]);
void check_redundancies(void);
void print_redundant_vertices(void);

Flip_info *flip_info(void);
void flip_hook(void (* routine) (char, const int[]));
void flip_list(int n);
void flip_list_put(int e);

void pr_triangle(int e);


extern char *regular_flip_crash_path;

extern int flip_test_flag, flip_proto_flag;


/* Statistics for SoS primitives */

extern int sos_pos3[5][5], sos_sphere[5][5];

#define SUCC_FLIP      0
#define UNSUCC_FLIP    1
#define SEARCH         2
#define STACK          2


/*------------------------------------------------------------------------*/
/*                            search.c                                    */

/* This record is used for reporting information about the number of
   infinite faces in the final triangulation. 				  */

typedef struct infinite_num_type
{
  int ext_e, ext_f, ext_t;
} Infinite_num;


/* This record is used for reporting information about the search
   time and history dag size.						  */

typedef struct search_info_type
{
  int hier_size;
  int height, not_inserted, in_tests;
  float avg_depth, avg_depth_ninsert;
} Search_info;


/* Search Parameters */

/* Type of search */

#define CASCADE_SEARCH    1
#define TRIDAG_SEARCH     2
#define NO_SEARCH         3

/* How often to check for redundancy */

#define redcheck_dist 	5

extern int Seed, in_tests, ivertex;

/* These next three variables are used by the different search methods */

extern long int tot_vis;
extern int not_inserted;
extern long vis_not_inserted;

/* These function pointers are used to specify the search method. */

extern  int (* search_make)(int i, int j, int k);
extern  int (* search_current_triangle)(int t);
extern void (* search_init)(int vertex[], int n);
extern  int (* search_locate_vertex)(int v);
extern  int (* search_locate_redundant)(int v);
extern void (* search_update_insert)(int v, int ijk, const int field[/*6*/], 
				     int i, int j, int k, int l);
#ifdef OLDSTUFF
extern void (* search_update_triangle)(int xyz, const int field[/*6*/], 
				       int x, int y, int z, int o, int p);
extern void (* search_update_edge)(int pox, int poy, int poz, 
				   const int field[/*6*/], int x, int y, 
				   int z, int o, int p);
extern void (* search_update_vertex)(const int field[/*9*/]);
#endif
extern void (* search_update_triangle)(int x, int y, int z, int o, int p,
				const int field[/*5*/]);
extern void (* search_update_edge)(int x, int y, int z, int o, int p,
			    const int field[/*6*/]);
extern void (* search_update_vertex)(const int field[/*9*/]);


extern void (* search_close)(void);

void search_open (int search_type);
int search_insert(int v);
int search_in_tetra(int i, int j, int k, int l, int v);
int search_pos_orientation(int v, int e);
int search_is_infinite_triangle(int e);
void kill_infinite_triangles(void);
Infinite_num infinite_num(Trist_num tq);
Search_info search_info(int n);
void print_search_dag(void);
int search_dag_height(void);

/*------------------------------------------------------------------------*/
/*                            tridag.c                                    */

/* This data structure is used to store the history dag.  The record is
   used as the data field in trist. */

typedef struct dag_record_type
{
  int search[2];             /* The next face to examine in the search --
                                -1 = done.                                 */
  Trist_org opposite[2];     /* The opposite vertex -- for regularity test */

#ifdef SDEBUG
  int flip_type;
#endif
} Dag_record;

#ifdef SDEBUG

/* flip types -- only used for debugging */
#define F_TRIANGLE 1
#define F_EDGE     2
#define F_VERTEX   3
#define F_INSERT   4

#endif

int pos_search(int t);
int neg_search(int t);
void set_pos_search(int t1, int t2);
void set_neg_search(int t1, int t2);
int search_ptr(int e);
void set_search_ptr(int e, int t);

Trist_org opposite_vertex(int e);
void set_opposite_vertex(int e, Trist_org v);
void set_flip_type(int t, int type), print_flip_type(int t);
int is_current_triangle (int t);

int tridag_make (int i, int j, int k);
void tridag_kill (void);

void tridag_init(int vertex[], int n);
int tridag_locate_vertex(int v);
int tridag_locate_redundant(int v);
void tridag_update_insert(int v, int ijk, const int field[/*9*/], 
			  int i, int j, int k, int l);
#ifdef OLDSTUFF
void tridag_update_triangle(int xyz, const int field[/*5*/], 
			    int x, int y, int z, int o, int p);
void tridag_update_edge(int pox, int poy, int poz, const int field[/*5*/],
			int x, int y, int z, int o, int p);
void tridag_update_vertex(const int field[/*9*/]);
#endif

void tridag_update_triangle (int x, int y, int z, int o, int p,
			     const int field[/*10*/]);
void tridag_update_edge     (int x, int y, int z, int o, int p,
			     const int field[/*10*/]);
void tridag_update_vertex   (const int field[/*10*/]);

/*------------------------------------------------------------------------*/
/*                             cascade.c                                  */

/* 
 * Structure used as the data field in the trist record.  This record 
 * stores information about what vertices are currently located in the
 * tetrahedron indicated by the (oriented) triangle.
 */

typedef struct cascade_record_type
{
  Basic_istaque_adt stack[2];
} Cascade_record;


void cascade_init(int vertex[], int n);
void cascade_close(void);

int cascade_make(int i, int j, int k);

int cascade_locate_vertex(int v);
int cascade_locate_redundant(int v);

void cascade_update_insert(int v, int ijk, const int field[/*6*/], 
			   int i, int j, int k, int l);
#ifdef OLSTUFF
void cascade_update_triangle(int xyz, const int field[/*6*/], 
			     int x, int y, int z, int o, int p);
void cascade_update_edge(int pox, int poy, int poz, const int field[/*6*/], 
			 int x, int y, int z, int o, int p);
void cascade_update_vertex(const int field[/*9*/]);
#endif

void cascade_update_triangle (int x, int y, int z, int o, int p, 
			      const int field[/*10*/]);
void cascade_update_edge     (int x, int y, int z, int o, int p, 
			      const int field[/*10*/]);
void cascade_update_vertex   (const int field[/*10*/]);

void cascade_print_locations(void);

/*------------------------------------------------------------------------*/
/*                              misc.c                                    */

void randomize (int v[], int n);
void load_point(int i, long x, long y, long z);
void scan_vertices (int n0, int *n1, int vertex[], int randomized, 
		    int proto_flag);
int lexi_compare(const int *i, const int *j);


/*--------------------------------------------------------------------------*/
/*                              stack.c                                     */

typedef char* Casc_Stack_adt;  /* Abstract data type! */

Casc_Stack_adt casc_stack_new (int minsize);
void           casc_stack_dispose (Casc_Stack_adt s);

void casc_stack_clear (Casc_Stack_adt s);
int  casc_stack_empty (Casc_Stack_adt s);

int  casc_stack_top (Casc_Stack_adt s);
int  casc_stack_pop (Casc_Stack_adt s);
void casc_stack_push (Casc_Stack_adt s, int value);
int  casc_stack_bot (Casc_Stack_adt s);
int  casc_stack_get (Casc_Stack_adt s);

int    casc_stack_length (Casc_Stack_adt s);
void   casc_stack_print (Casc_Stack_adt s, FILE *file);
double casc_stack_resize (Casc_Stack_adt s, double resize_code);

int casc_max_storage(void);

#endif  /* #ifndef __DELCX_H__ */




