/* delcx/flip.c --- "flips" 3-d triangulation to DT or Regular Triang. */
/* Revised -- 4/14/92 -- Includes combined trist and search data structures */

/* compile with cc -DEBUG for debugging and to use flags */

#include "delcx.h"
#include "dt.h"
#include "version.h"
#include "primflip.h"

/* #define REPORT_THE_CRASH 1 */


/*--------------------------------------------------------------------------*/

/* flags and proto macros --- for debugging only */

#ifdef __DEBUG__
int flip_test_flag = FALSE;
int flip_proto_flag = FALSE;
#define proto(PRINT_COMMAND)  if (flip_proto_flag) PRINT_COMMAND
#else
#define proto(PRINT_COMMAND)  /* do nothing */
#endif

/*--------------------------------------------------------------------------*/

static int next_nd(void);
static void add_nd(int e);

static void set_white(int t), set_green(int t), set_red(int t);
static void check_triangle(int e);

static int concave_edge(int e); 
static int flippable_edge(int e);
static int check_ev_flip(int e); 

static int valid_4to1_flip(int e);
static int valid_4to1_pos(int e, int vert[]);
static int check_ring(int t, int i, int j, int k, int l, int m);

static int flip_triangle(int xyz);
static int flip_edge(int pox);
static int flip_vertex(int e);

static void diagonalize(int xyz, int field[/*6*/], 
			int x, int y, int z, int o, int p);
static int diagonalize_test (int xpo, int x, int y, int z, int p, int o);
static void cut(int field[/*6*/], int x, int y, int z, int o, int p);
static void load(int xyo, int x, int y, int z, int o, int p, 
		 int field[/*6*/]);
static void load2(int e, /* i, j, k, l, m, */ int field[/*9*/]);


static void emergency_break(void);
static void crash(void);

static int ch_triangle(int e, int i, int j, int k);

static void Fmerge(int d, int e);


static Basic_istaque_adt nd;

#define NONE -666

#define _WHITE  ' '  /* Regular */
#define _GREEN  'g'  /* non-Regular */
#define _RED    'r'  /* non-transformable */
#define _BLUE   'b'  /* unknown */

static void (* hook) (char, const int[]) = NULL;

static int reds, greens;
static int skips = 0, eflips = 0, tflips = 0, vflips = 0;
static int flipc;

static int mainloops = 0;  /* see emergency_break() */

int regular_triang = TRUE;

char *regular_flip_crash_path;

#define is_hull_triangle(T)		((T) < 5)

/*--------------------------------------------------------------------------*/

/* Variables for flipped vertices */

Trist_org * flipped_v;
int * flipped_loc;
int max_flipped;


/*--------------------------------------------------------------------------*/

/* Statistics for SoS primitives */
/* first index -- #infinite pts; second -- source */

int sos_pos3[5][5] = { {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0} };
int sos_sphere[5][5] = { {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0}, 
			 {0, 0, 0, 0, 0} };
int test1, test2, test3; /* number of infinite points in tests */

/*------------------------------------------------------------------------*/

/*
 * open_flip (n) -- This function opens the flip module by initializing
 *		    various global variables.  n is the number of
 *		    vertices.  Note that the triangulation must at least
 *		    include a tetrahedron.
 */
void open_flip (int n)
{
#ifdef REGULAR
  max_flipped = n - 4;
  flipped_v = MALLOC (Trist_org, 1 + max_flipped);
#endif
  /*
  flip_test_flag = TRUE;
  flip_proto_flag = TRUE;
  */
}


/*------------------------------------------------------------------------*/

/*
 * close_flip () -- This function closes the flip module by deallocating
 *	 	     various global variables.
 */
void close_flip (void)
{
#ifdef REGULAR
  FREE (flipped_v);
  if (flipped_loc)
    FREE (flipped_loc);
#endif
}

/*-----------------------------------------------------------------------*/

/*
 * construct_outer_tetra () -- Initializes the tetrahedra which will 
 *            contain all of the points in the point set.
 */
void construct_outer_tetra (void)
{
  int jik, ijk, jki, kij, ijv, jkv, kiv;

  /* i,j,k, and v are the last four vertices */
  int i = num_vert + 1, j = num_vert + 2;
  int k = num_vert + 3, v = num_vert + 4;

#ifdef SDEBUG
  print("Initializing the outer triangle\n");
#endif

  /* Load the outer vertices */
  load_point(i,   1,   0,   1);
  load_point(j, - 1,   0,   1);
  load_point(k,   0,   1, - 1);
  load_point(v,   0, - 1, - 1);


  /* Initialize the SoS package for infinite points */

#ifdef SDEBUG
  printf("\n\nTesting outer tetrahedron:\n");
  printf("Orientation of (%d,%d,%d,%d) = %d\n", i, j, k, v,
	 sos_positive3_inf(i,j,k,v));
  printf("Orientation of (%d,%d,%d,%d) = %d\n", j, i, k, v, 
	 sos_positive3_inf(j,i,k,v));
#endif

  jik = search_make (j, i, k);
  ijk = Sym(jik);
  jki = Enext(ijk);
  kij = Enext(jki);

  ijv = search_make(i, j, v);
  jkv = search_make(j, k, v);
  kiv = search_make(k, i, v);

  /* test all triangles before splicing */
  (void) dt_test_triangle (ijk, i, j, k);
  (void) dt_test_triangle (ijv, i, j, v);
  (void) dt_test_triangle (jkv, j, k, v);
  (void) dt_test_triangle (kiv, k, i, v);

  Fsplice (ijk, ijv);
  Fsplice (jki, jkv);
  Fsplice (kij, kiv);
  Fsplice (Enext (ijv), Enext (Sym (jkv)));
  Fsplice (Enext (jkv), Enext (Sym (kiv)));
  Fsplice (Enext (kiv), Enext (Sym (ijv)));

  trist_hull_facet_set (Sym(ijk), 1);
  trist_hull_facet_set (ijv, 1);
  trist_hull_facet_set (jkv, 1);
  trist_hull_facet_set (kiv, 1);

  (void) dt_test_open_tetra(ijv, i, j, k, v);
}

/*--------------------------------------------------------------------------*/

/*
 * flip_transform() -- This function uses a stack to search for non-regular
 *                     link facets, and then flips them when possible.
 */
int flip_transform (void)
     /* flips given trist until all triangles locally regular;
        all possibly non-regular triangles are assumed to put on nd list
        with flip_list_put()  (user needs to call flip_list() first!)
        returns number of flips (flipc). */
{
  int e;
  Assert_always (nd);
  Assert (greens == basic_istaque_length (nd));
  flipc = 0;
  reds = 0;
  while ((e = next_nd ()) != NONE)
    {
      emergency_break ();
      if (not check_ev_flip(e))		/* Check for edge or vertex flip */
	if (not flip_triangle (e)) {     /* If not, non-transformable */
	    skips ++;
	    set_red (TrIndex (e));
	  }
    }
  if (reds or greens)
    crash ();
  return (flipc);
}

/*--------------------------------------------------------------------------*/

/*
 * next_nd () -- This function finds the next non-regular link facet.
 */
static int next_nd (void)
     /* returns next _GREEN triangle from nd list, or NONE */
{
  int t, i;
  while (not basic_istaque_empty (nd))
    {
      i = basic_istaque_pop (nd);  /* stack */
      t = TrIndex (i);
      if ((is_current_triangle (t)) and (trist_color (t) == _GREEN))
        return (i);  /* only if still unsolved and not vanished */
    }
  return (NONE);
}

/*--------------------------------------------------------------------------*/

static void add_nd (int e)
{
  basic_istaque_push (nd, e);
  greens ++;
}

/*--------------------------------------------------------------------------*/
/* FUNCTIONS FOR SETTING TRIANGLE COLOR 				    */
/*--------------------------------------------------------------------------*/

static void set_white (int t)
{
  char c = trist_color (t);
  if (c == _RED)
    reds --;
  else if (c == _GREEN)
    greens --;

  trist_color_set (t, _WHITE);
}

/*--------------------------------------------------------------------------*/

static void set_green (int t)
{
  Assert (trist_color (t) != _GREEN);

  if (trist_color (t) == _RED)
    reds --;
  trist_color_set (t, _GREEN);
}

/*--------------------------------------------------------------------------*/

static void set_red (int t)
{
  if (trist_color (t) == _GREEN)
    greens --;
  trist_color_set (t, _RED);
  reds ++;
}

/*--------------------------------------------------------------------------*/

void set_blue (int t)
{
  int c = trist_color(t);
  Assert_always ((c == _WHITE) or (c == _BLUE));
  trist_color_set (t, _BLUE);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static void check_triangle (int e)
     /* must be called "BEFORE" checking the edges */
{
  int t = TrIndex (e);
  char c = trist_color (t);

  if (is_regular (e))
    { /* note: it might be a hull triangle, ie. c == '*', or already _WHITE */
      if ((c == _RED) or (c == _GREEN) or (c == _BLUE))
	set_white (t);
    }
  else if (c != _GREEN)
      {
	set_green (t);
	add_nd (e);
      }
}

/*--------------------------------------------------------------------------*/

/*
 * delete_the_triangle () -- This function removes triangle e from the
 *                           triangulation without deleting the triangle
 *                           record.
 */
void delete_the_triangle (int e)
{
  register int d = Enext (e), f = Enext2 (e);

#ifdef VDEBUG
  print ("     Deleting triangle ");
  pr_triangle(e);
#endif
  set_white (TrIndex(e));    /* updates the counts of reds and greens */

  Fsplice (Sym (Fnext (Sym (e))), e);
  Fsplice (Sym (Fnext (Sym (d))), d);
  Fsplice (Sym (Fnext (Sym (f))), f);
}

/*--------------------------------------------------------------------------*/

int is_regular (int xyz)
     /* tests for local regularity */
{
  int x, y, z, o, p;
  int temp;

  if (is_hull_triangle (TrIndex (xyz)))
    return (TRUE);
  trist_triangle (xyz, &x, &y, &z);
  o = Dest (Enext (Fnext (xyz)));
  p = Dest (Enext (Fnext (Sym (xyz))));
  Assert_if (flip_test_flag,
             (    sos_positive3_inf (o, x, y, z)
              and (not sos_positive3_inf (p, x, y, z))
              and dt_test_triangle (Fnext (xyz), x, y, o)
              and dt_test_triangle (Fnext (Sym (xyz)), y, x, p)));
  
  return (sos_in_sphere_inf_p (o, x, y, z, p));
}

/*--------------------------------------------------------------------------*/

/*
 * concave_edge (e) -- Tells whether the edge ij of triangle e = ijk is
 *                     concave with the following criterion:  If there are
 *                     three faces incident to e, none of which are hull 
 *                     facets, then e is concave.
 */
static int concave_edge(int e)
{
  int e1, e2;

  e1 = Fnext(e);
  e2 = Fnext(e1);

  return (    (Fnext(e2) == e)
	  and (not is_hull_triangle(TrIndex(e1))) 
	  and (not is_hull_triangle(TrIndex(e2))));
}


/*--------------------------------------------------------------------------*/

/*
 * flippable_edge (e) -- Checks whether edge e is flippable as determined
 *                       by its orientation.  This is only used for the
 *                       regular triangulation.
 */
static int flippable_edge(int e)
{
  int e1, e2;
  int x, y, z, o, p;
  int val;

  e1 = Fnext(e);
  e2 = Fnext(e1);
  y = Dest(Enext(e1));
  z = Dest(Enext(e2));
  trist_triangle(e, &p, &o, &x);
#ifdef DEBUG
  print ("Testing position with xyz = %d, %d, %d; o = %d, p = %d\n",
	 x, y, z, o, p);
#endif

  val = (sos_positive3_inf(o, x, y, z) != sos_positive3_inf(p, x, y, z));
  if (val) {
    sos_pos3[ num_infinite4(o, x, y, z) ][SUCC_FLIP] ++;
    sos_pos3[ num_infinite4(p, x, y, z) ][SUCC_FLIP] ++; 
  }
  else {
    sos_pos3[ num_infinite4(o, x, y, z) ][UNSUCC_FLIP] ++;
    sos_pos3[ num_infinite4(p, x, y, z) ][UNSUCC_FLIP] ++;
  }

  return (val);
}

/*------------------------------------------------------------------------*/

/*
 * check_ev_flip(e) -- Check for an edge or vertex flip for triangle e.
 *                     If an edge has three incident triangles, none of
 *                     which are hull facets, then it is concave.
 */
static int check_ev_flip(int e)
{
  int conc, conc1, conc2;    /* concavity of three edges of e */
  int num_conc, flipt;

  conc = concave_edge(e);
  conc1 = concave_edge(Enext(e));
  conc2 = concave_edge(Enext2(e));

  num_conc = (conc ? 1 : 0) + (conc1 ? 1 : 0) + (conc2 ? 1 : 0);

  if (num_conc == 1) {
    if (conc)
      flipt = e;
    else if (conc1)
      flipt = Enext(e);
    else
      flipt = Enext2(e);
#ifdef REGULAR
    /* If this edge is not flippable, then TRUE is returned because no
       other flip (i.e. vertex or triangle flip) can be performed. */

    if (regular_triang)
      if (not flippable_edge (flipt)) {
	skips ++;
	set_red(TrIndex(e));
	return (TRUE);
      }
#endif
    flip_edge(flipt);
    return (TRUE);
  }
  else {
#ifdef REGULAR
    if (regular_triang) {
      if (num_conc == 2) {
	/* find flippable vertex */
	if (not conc)
	  flipt = Enext(e);
	else if (not conc1)
	  flipt = Enext2(e);
	else
	  flipt = e;
	if (not flip_vertex(flipt)) {
	  skips ++;
	  set_red(TrIndex(e));
	}
	return (TRUE);
      }
      else {
	return(FALSE);
      }
    }
    else {
      if (num_conc == 2) {
	skips ++;
	set_red(TrIndex(e));
      }
      return(num_conc == 2);    /* If this is true, a triangle flip cannot
				   be performed. */
    }
#else
    if (num_conc == 2) {
      skips ++;
      set_red(TrIndex(e));
    }
    return(num_conc == 2);
#endif
  }
}


/*--------------------------------------------------------------------------*/

/*
 * valid_4to1_flip (e) --
 *        Determines if the destination of triangle e can be flipped out
 *        of the triangulation.  No assumptions are made about the
 *        positions of t and surrounding triangles.
 */
static int valid_4to1_flip (int e)
{
  int vert[5];
  int retval = FALSE;

  if (valid_4to1_pos (e, vert)) {
    /* vert[1] is the point inside the convex hull of the other four points */
    retval = (not sos_in_sphere_inf_p (vert[0], vert[2], vert[3], 
				   vert[4], vert[1]));
    in_tests ++;
  }
  return (retval);
}

/*--------------------------------------------------------------------------*/

/*
 * valid_4to1_pos (e, vert) --
 *        Determines if the destination of triangle e is in the correct
 *        position relative to surrounding points for a 4 to 1 flip to
 *        be performed.
 */
static int valid_4to1_pos (int e, int vert[])
{
  int e1 = Fnext(e);
  int e2 = Fnext(e1);
  int i, j, k, l, m;     /* The five vertices involved in the 
			    four tetrahedra. */
  int dummy1, dummy2;
  int retval;

#ifdef VDEBUG
  print ("\n\nThe triangles:\n");
  pr_triangle(e);
  pr_triangle(e1);
  pr_triangle(e2);
  print ("Destination of e is %d\n", Dest(e));
#endif

  if (Fnext(e2) != e) {
#ifdef VDEBUG
    print ("Doesn't have 3 triangles!\n");
#endif
    return (FALSE);
  }

  trist_triangle (e, &i, &j, &k);
  trist_triangle (e1, &dummy1, &dummy2, &l);
  trist_triangle (e2, &dummy1, &dummy2, &m);

  /* Load the vertices into the vert array.  These will be used by
     the caller.  j is the inner point, so it is placed last.       */

  vert[0] = i; vert[1] = j; vert[2] = k; vert[3] = l; vert[4] = m;

  retval =    check_ring (Sym(Enext(e)),  k, j, i, m, l)
	  and check_ring (Sym(Enext(e1)), l, j, i, k, m)
	  and check_ring (Sym(Enext(e2)), m, j, i, l, k);

#ifdef VDEBUG
  if (retval)
    print ("\n\n&&&&&&&&&&&&&&&& Flipping vertex %d\n\n", vert[1]);
#endif

  return (retval);
}

/*--------------------------------------------------------------------------*/

/*
 * check_ring (t, i, j, k, l, m) --
 *        Checks that the face ring containing t has only 3 faces, and
 *        that the faces correspond to the correct faces.
 */
static int check_ring (int t, int i, int j, int k, int l, int m)
{
  int t1, t2;
  int ret_val;

#ifdef VDEBUG
  print ("Vertices: %d, %d, %d, %d, %d\n", i, j, k, l, m);
#endif

  t1 = Fnext(t);
  t2 = Fnext(t1);

#ifdef VDEBUG
  print ("The triangles:\n");
  pr_triangle(t);
  pr_triangle(t1);
  pr_triangle(t2);
#endif

  if ((t != Fnext(t2)) or (t == t1) or (t1 == t2) or (t2 == t))
    return(FALSE);

  ret_val = (    (ch_triangle(t,  i, j, k))
	     and (ch_triangle(t1, i, j, l))
	     and (ch_triangle(t2, i, j, m)));
  return (ret_val);
}

/*----------------------------------------------------------------------*/

/*
 * flip_insert (ijk, v) -- Insert vertex v inside the tetrahedra containing
 *                         triangle ijk.  The four new tetrahedra are 
 *                         are created, and the search data structure is 
 *                         updated.
 */
void flip_insert (int ijk, int v)
{
  int i, j, k, l;
  int jil, kjl, ikl, ijv, jkv, ikv, ilv, jlv, klv;
  int t;                                /* The root of the search tree */
  int field[9];				/* Used for passing the triangles
                                           to the update procedure.       */

  ivertex = v;

  /* Find the other three faces of the tetrahedra */

  jil = Sym (Fnext (ijk));
  kjl = Sym (Fnext (Enext (ijk)));
  ikl = Sym (Fnext (Enext2 (ijk)));

  /* Find the vertex numbers */

  trist_tetra(ijk, &i, &j, &k, &l);

#ifdef SDEBUG
  print ("Testing that point is inside tetrahedron\n");
  print ("(%d, %d, %d, %d) = %d\n", v, i, j, k, sos_positive3_inf (v,i,j,k));
  print ("(%d, %d, %d, %d) = %d\n", v, j, i, l, sos_positive3_inf (v,j,i,l));
  print ("(%d, %d, %d, %d) = %d\n", v, k, j, l, sos_positive3_inf (v,k,j,l));
  print ("(%d, %d, %d, %d) = %d\n", v, i, k, l, sos_positive3_inf (v,i,k,l));

  Assert_always (    sos_positive3_inf (v,i,j,k)
		 and sos_positive3_inf (v,j,i,l)
		 and sos_positive3_inf (v,k,j,l)
		 and sos_positive3_inf (v,i,k,l));
#endif

#ifdef SDEBUG
  printf("Inserting vertex %d in tetrahedra (%d, %d, %d, %d).\n",
	  v, i, j, k, l);
#endif

  /* Create the 6 new triangles that are needed */

  ijv = search_make(i, j, v);
  jkv = search_make(j, k, v);
  ikv = search_make(i, k, v);
  ilv = search_make(i, l, v);
  jlv = search_make(j, l, v);
  klv = search_make(k, l, v);

  t = TrIndex (jlv);

  /* Update the face rings for the tetrahedra faces */

  Fmerge(ijk, ijv);
  Fmerge(Enext(ijk), jkv);
  Fmerge(Enext2(ijk), Sym(ikv));
  Fmerge(Enext(jil), ilv);
  Fmerge(Enext(ikl), klv);
  Fmerge(Enext(kjl), jlv);

  /* Create the rings containing only the new faces */

  /* vij -> vil -> vik */

  Fmerge(Enext2(ijv), Enext2(ilv));
  Fmerge(Enext2(ilv), Enext2(ikv));

  /* vjk -> vjl -> vji */

  Fmerge(Enext2(jkv), Enext2(jlv));
  Fmerge(Enext2(jlv), Sym(Enext(ijv)));

  /* kvi -> kvj -> kvl */

  Fmerge(Enext(ikv), Enext(jkv));
  Fmerge(Enext(jkv), Enext2(Sym(Enext(klv))));

  /* lvi -> lvk -> lvj */

  Fmerge(Enext(ilv), Enext(klv));
  Fmerge(Enext(klv), Enext(jlv));

  /* Now, test all of the tetrahedra that were created */

#ifdef SDEBUG
  dt_test_triangle(ijv, i, j, v);
  dt_test_triangle(ikv, i, k, v);
  dt_test_triangle(ilv, i, l, v);
  dt_test_triangle(jkv, j, k, v);
  dt_test_triangle(jlv, j, l, v);
  dt_test_triangle(klv, k, l, v);
  dt_test_open_tetra(ijv, i, j, k, v);
  dt_test_open_tetra(Sym(ijv), j, i, l, v);
  dt_test_open_tetra(Sym(jkv), k, j, l, v);
  dt_test_open_tetra(ikv, i, k, l, v);
#endif

  /* We also need to update the search data structure */

/*
  print ("ijv = %d, %s\n", TrIndex(ijv), 
	 (Odd(TrVersion(ijv)) ? "odd" : "even"));
  print ("jkv = %d, %s\n", TrIndex(jkv), 
	 (Odd(TrVersion(jkv)) ? "odd" : "even"));
  print ("ikv = %d, %s\n", TrIndex(ikv), 
	 (Odd(TrVersion(ikv)) ? "odd" : "even"));
  print ("ilv = %d, %s\n", TrIndex(ilv), 
	 (Odd(TrVersion(ilv)) ? "odd" : "even"));
  print ("jlv = %d, %s\n", TrIndex(jlv), 
	 (Odd(TrVersion(jlv)) ? "odd" : "even"));
  print ("klv = %d, %s\n", TrIndex(klv), 
	 (Odd(TrVersion(klv)) ? "odd" : "even"));
*/

  field[0] = ijv; field[1] = ikv; field[2] = jkv;
  field[3] = ilv; field[4] = jlv; field[5] = klv;
  field[6] = jil; field[7] = kjl; field[8] = ikl;

  search_update_insert (v, ijk, field, i, j, k, l);

  /* Finally, put all the triangles onto the flip list. */

  flip_list_put(ijk);
  flip_list_put(jil);
  flip_list_put(kjl);
  flip_list_put(ikl);
  set_blue (TrIndex(jlv));
  set_blue (TrIndex(ilv));
  set_blue (TrIndex(klv));
  set_blue (TrIndex(ikv));
  set_blue (TrIndex(ijv));
  set_blue (TrIndex(jkv));
}

/*--------------------------------------------------------------------------*/

/*
 * flip_triangle (xyz) -- attempts to flip the triangle xyz.  Assumes that
 *                        xyz is a link facet, and that p_i, the inserted
 *                        point is on the positive side of xyz.  This means
 *                        that o in the code below is p_i.
 */
static int flip_triangle (int xyz)
     /* assumes is_regular (xyz) == FALSE */
{
  int x, y, z, o, p, xyo = Fnext (xyz), j, field[10];
  int concave;
  trist_triangle (xyz, &x, &y, &z);
  o = Dest (Enext (xyo));
  p = Dest (Enext (Fnext (Sym (xyz))));

  Assert_always (o == ivertex);

  /* NOTE: test1-3 in the code below is used for statistical purposes */

  concave = TRUE;
  test1 = num_infinite4(p, y, z, o);
  if (not sos_positive3_inf (p, y, z, o)) {
    test2 = num_infinite4(p, z, x, o);
    if (not sos_positive3_inf (p, z, x, o)) {
      test3 = num_infinite4(p, x, y, o);
      if (not sos_positive3_inf (p, x, y, o))
	concave = FALSE;
    }
    else
      test3 = 4;
  }
  else
    test2 = test3 = 4;        /* means that these tests weren't made */

  proto (print ("%s [%d] t%d = %d,%d,%d;%d,%d\n",
                If((concave), "SKIP tetra", "FLIP tri to edge"),
                xyz, TrIndex (xyz), x, y, z, p, o));
  if (concave) {
    /* update stats */
    sos_pos3[ test1 ][UNSUCC_FLIP] ++;
    sos_pos3[ test2 ][UNSUCC_FLIP] ++;
    sos_pos3[ test3 ][UNSUCC_FLIP] ++;

    return (FALSE);
  }
  else
    { /* flip: triangle to edge;
         assuming sos_positive3 (o, x, y, z) and sos_positive3 (p, y, x, z) */

      /* update stats */
      sos_pos3[ test1 ][SUCC_FLIP] ++;
      sos_pos3[ test2 ][SUCC_FLIP] ++;
      sos_pos3[ test3 ][SUCC_FLIP] ++;

      tflips ++;
      flipc ++;

      prim_flip_triangle (xyz, field);

      set_white(TrIndex(xyz));
      upfor (j, 4, 6)
	set_blue(TrIndex(field[j]));
      upfor (j, 7, 9)               /* only link facets are put on the stack */
        check_triangle (field[j]);
      if (hook)
	hook (FLIP_TRIANGLE, field);

      search_update_triangle (x, y, z, o, p, field);

      return (TRUE);
    }
}

/*--------------------------------------------------------------------------*/

/*
 * flip_edge (pox) --     attempts to flip the edge op.  Assumes that
 *                        pox is a link facet, and that p_i, the inserted
 *                        point is on the positive side of pox.  This means
 *                        that y in the code below is p_i.
 */
static int flip_edge (int pox)
{
  int x, y, z, o, p, poy, poz, xyo, j, field[10];
  poy = Fnext (pox);
  poz = Fnext (poy);
  trist_triangle (pox, &p,  &o,  &x);
  y = Dest (Enext (poy));
  z = Dest (Enext (poz));

  Assert_always (y == ivertex);

  proto (print ("FLIP edge to tri [%d|%d|%d] :: %d,%d,%d;%d,%d\n",
                pox, poy, poz, x, y, z, o, p));
  /* flip edge to triangle;
     assuming sos_positive3 (o, x, y, z) and sos_positive3 (p, y, x, z) */
  eflips ++;
  flipc ++;

  prim_flip_edge (pox, field);

  set_white(TrIndex(pox));
  set_white(TrIndex(poy));
  set_white(TrIndex(poz));

  /* Only link facets are checked */
  check_triangle (Enext(field[8]));         /* xzo */
  check_triangle (Enext2(field[9]));        /* zxp */
  set_blue (TrIndex(field[4]));
  set_blue (TrIndex(field[5]));
  set_blue (TrIndex(field[6]));
  set_blue (TrIndex(field[7]));

  if (hook)
    hook (FLIP_EDGE, field);

  search_update_edge (x, y, z, o, p, field);

  return (TRUE);
}

/*--------------------------------------------------------------------------*/

#ifdef REGULAR
/*
 * flip_vertex (ef) --
 *        This procedure will attempt to flip the vertex that is the
 *        destination of triangle ef.
 */
static int flip_vertex (int ef)
{
  int i, field[10];     /* All the triangles involved in this flip */
  int v = Dest(ef);     /* The flipped vertex */
  int w, x, y, z;

  if (not valid_4to1_flip (ef))
    {
/*      print ("\n\n**** NOT A VALID 4:1 FLIP!!! ****\n\n"); */
      return (FALSE);
    }

#ifdef VDEBUG
  print ("&&&& Performing vertex flip with vertex %d\n", Dest(ef));
#endif

  flipc++;

  prim_flip_vertex (ef, field);

  upfor (i, 0, 5)
    set_white (TrIndex(field[i]));

  /* field[8] is the only new link facet */
  flip_list_put (field[8]);
  add_redundant_vertex(Dest(ef));

  if (hook)
    hook (FLIP_VERTEX, field);

  search_update_vertex (field);

  return (TRUE);
}
#endif


/*--------------------------------------------------------------------------*/

/*
 * diagonalize () and cut ()  -- This two functions are refinements of
 *        flip_triangle and flip_edge respectively.  They perform the
 *        trist manipulations necessary to update the triangle
 *        structure.
 */

#ifndef __DEBUG__
/*ARGSUSED*/
#endif
static void diagonalize (int xyz, int field[/*6*/], 
			 int x, int y, int z, int o, int p)
     /* assumes that we have a double tetrahedron;
         creates 3 new triangles, i.e, 3*3 edfacets   */
{
  int xpo, pyo, pzo, oxz, zpx, xpz, zox, yoz, pyz;
  xpo = search_make (x, p, o);
  pyo = search_make (p, y, o);
  pzo = search_make (p, z, o);
  
  oxz = Sym (Enext (field[2]));
  zpx = Enext (field[4]);
  xpz = Sym (Enext (zpx));
  zox = Enext2 (oxz);
  yoz = Enext (Sym (field[1]));
  pyz = Sym (Enext (field[5]));
  Fsplice (xpz, xpo);
  Fsplice (oxz, Enext2 (xpo));
  Fsplice (pyz, pyo);
  Fsplice (yoz, Enext (pyo));
  Fsplice (Sym (zpx), pzo);
  Fsplice (zox, Enext (pzo));
  Fsplice (Enext (xpo), Enext (Sym (pyo)));
  Fsplice (Enext (Sym (pyo)), Enext (Sym (pzo)));
  Assert_if (flip_test_flag, diagonalize_test (xpo, x, y, z, p, o));
}

/*--------------------------------------------------------------------------*/

#if __DEBUG__
static int diagonalize_test (int xpo, int x, int y, int z, int p, int o)
{
  int one, two, three;
  one   = Enext (xpo);
  two   = Fnext (one);
  three = Fnext (two);
  Assert_always (    (one == Fnext (three))
                 and dt_test_triangle (one,   p, o, x)
                 and dt_test_triangle (two,   p, o, y)
                 and dt_test_triangle (three, p, o, z)
                 and dt_test_open_tetra (one,   p, o, z, x)
                 and dt_test_open_tetra (two,   p, o, x, y)
                 and dt_test_open_tetra (three, p, o, y, z));
  return (TRUE);
}
#endif

/*--------------------------------------------------------------------------*/

#ifndef __DEBUG__
/*ARGSUSED*/
#endif
static void cut (int field[/*6*/], int x, int y, int z, int o, int p)
     /* assumes that we have a double tetrahedron;
         creates 1 new triangle, i.e, 1*3 edfacets + syms !! (every else too)*/
{
  int xyz;
  xyz = search_make (x, y, z);
  
  Fsplice (Sym (Fnext (Sym (field[0]))), xyz);
  Fsplice (Sym (Fnext (Sym (field[1]))), Enext (xyz));
  Fsplice (Sym (Fnext (Sym (field[2]))), Enext (Enext (xyz)));
  Assert_if (flip_test_flag,
             (    dt_test_open_tetra (Fnext (xyz), x, y, z, o)
              and dt_test_open_tetra (Fnext (Sym (xyz)), y, x, z, p)
              and dt_test_triangle (Fnext (Fnext (Sym (field[0]))), y, x, p)));
}

/*--------------------------------------------------------------------------*/

/******************** Procedures for flipped vertices ***********************/
/*
 * Flipped vertices are recorded in the flipped_v array.  The function
 * add_redundant_vertex is used to record a vertex.  check_vflips is used
 * to verify that each point is indeed redundant.  print_redundant_vertices
 * simply prints out a list of these vertices.
 */

#ifdef REGULAR

void add_redundant_vertex(Trist_org v)
{
  Assert_always(v);

  flipped_v[ ++vflips ] = v;
}

/*-------------------------------------------------------------------------*/

/*
 * compute_redundant_loc () -- This function will compute the containing 
 *                             tetrahedra for each redundant vertex.
 */
void compute_redundant_loc (void)
{
  int cnt;

  if (not flipped_loc)
    {
      flipped_loc = MALLOC (int, vflips + 1);

      for (cnt = 1; cnt <= vflips; cnt++)
	flipped_loc[cnt] = search_locate_redundant (flipped_v[cnt]);
    }
}

/*-------------------------------------------------------------------------*/

/*
 * update_redundant_loc () -- This function updates the location of the
 *                            redundant vertices when a permutation is
 *                            executed.  It is passed to the hook for
 *                            trist_permute ().
 */
int update_redundant_loc (const int map[])
{
  int cnt, tri;

  for (cnt = 1; cnt <= vflips; cnt++)
    {
      tri = flipped_loc[cnt];
      flipped_loc[cnt] = EdFacet (map[TrIndex(tri)], TrVersion(tri));

    }

}

/*-------------------------------------------------------------------------*/

/*
 * output_redundant_loc () -- This function stores information about the
 *                            redundant vertices in a binary file.  The
 *                            following information is stored:
 *
 *                                  1.  # of redundant vertices
 *                                  2.  flipped_v array
 *                                  3.  flipped_loc array
 */
void output_redundant_loc(char data_name[])
{
  char *path_name = STRDUP (basic_cb_frmt ("%s.red", data_name));
  FILE *f = basic_fopen(path_name, "w");

  print ("Saving redundant vertex locations to \"%s\" ...\n", path_name);

#ifdef DEBUG
  {
    int cnt, tri, i, j, k, l, v;

    print ("\n");
    upfor (cnt, 1, vflips)
      {
	v = flipped_v[cnt];
	tri = flipped_loc[cnt];
	trist_tetra (tri, &i, &j, &k, &l);
	print ("%d: ", v);
	if (not (    sos_positive3_inf (v,i,j,k)
	         and sos_positive3_inf (v,j,i,l)
		 and sos_positive3_inf (v,k,j,l)
		 and sos_positive3_inf (v,i,k,l)))
	  basic_error ("Redundant vertex not in correct tetrahedron!\n");
      }
  }
#endif

  binfwrite (&vflips, 1, f);
  binfwrite (flipped_v, vflips + 1, f);
  binfwrite (flipped_loc, vflips + 1, f);
  basic_fclose (f);
}

/*-------------------------------------------------------------------------*/

/* NOTE:  THIS FUNCTION SHOULD BE MOVED SOMEWHERE ELSE WHERE IT CAN BE
          INCLUDED IN OTHER APPLICATIONS, PROBABLY IN ALF LIBRARY!         */

/*
 * input_redundant_loc () -- This function reads information about the
 *                           redundant vertices from a binary file.  The
 *                           following information is stored:
 *
 *                                  1.  # of redundant vertices
 *                                  2.  flipped_v array
 *                                  3.  flipped_loc array
 */
Redundant_info *input_redundant_loc(char data_name[])
{
  char *path_name = STRDUP (basic_cb_frmt ("%s.red", data_name));
  FILE *f = basic_fopen(path_name, "r");

  Redundant_info *ri = MALLOC (Redundant_info, 1);

  print ("Reading redundant vertex locations from \"%s\" ...\n", path_name);

  binfread (&(ri->vflips), 1, f);
  ri->flipped_vert = MALLOC (Trist_org, ri->vflips + 1);
  ri->flipped_loc = MALLOC (int, ri->vflips + 1);
  binfread (ri->flipped_vert, vflips + 1, f);
  binfread (ri->flipped_loc, vflips + 1, f);
  basic_fclose(f);

#ifdef DEBUG
  {
    int cnt, tri, i, j, k, l, v;

    print ("\n");
    upfor (cnt, 1, ri->vflips)
      {
	v = ri->flipped_vert[cnt];
	tri = ri->flipped_loc[cnt];
	trist_tetra (tri, &i, &j, &k, &l);
	if (not (    sos_positive3_inf (v,i,j,k)
	         and sos_positive3_inf (v,j,i,l)
		 and sos_positive3_inf (v,k,j,l)
		 and sos_positive3_inf (v,i,k,l)))
	  basic_error ("Redundant vertex %d not inside!\n", v);
      }
  }
#endif

  return (ri);
}

/*-------------------------------------------------------------------------*/

/*
 * check_redundancies () -- This function will check to make sure that
 *                          redundant vertices are actually redundant by
 *                          locating the containing tetrahedron in the
 *                          final triangulation and checking for
 *                          redundancy.
 */
void check_redundancies (void)
{
  int cnt, tri;
  int i,j,k,l, d1, d2;

  print ("Checking redundant vertices:  ");
  for (cnt = 1; cnt <= vflips; cnt++) {
    print ("%d ", flipped_v[cnt]);
    tri = flipped_loc[cnt];
    trist_triangle (tri, &i, &j, &k);
    trist_triangle (Fnext(tri), &d1, &d2, &l);
    Assert_always (    search_in_tetra (i, j, k, l, flipped_v[cnt])
		   and (not sos_in_sphere_inf (i, j, k, l, 
				       flipped_v[cnt])));
    if (cnt % 30 == 0)
      print ("\n");
  }
  print ("\n");
}

/*-------------------------------------------------------------------------*/

void print_redundant_vertices (void)
{
  int index;

  print ("\n\nRedundant vertices:\n");

  upfor (index, 1, vflips) {
    if ((index % 20) == 0)
      print ("\n");
    print ("%d ", flipped_v[index]);
  }
}
	
#endif

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

static void emergency_break (void)
{
  static int n, max;
  if (mainloops == 0)
    { /* initialize */
      n = trist_max_org ();
      max = 3   *   n * (n + 3) + 1;
      /* BIG NOTE:  There shouldn't be more than n * (n + 3) + 1
         flips (see Herbert's Triangulation notes, page 95), but
         this doesn't count the "skips" !!!
         (eg, file d/kurt)
         !!!!!!  THIS NEEDS TO BE FIXED/CLARIFIED   !!!!!!!!!!!!!!!! */
    }
  mainloops ++;
  Assert_always ((mainloops <= max) or (max <= 0));  /* emergency break */
}

/*--------------------------------------------------------------------------*/

Flip_info *flip_info (void)
{
  static Flip_info ti;

  ti.skips = skips;
  ti.eflips = eflips;
  ti.tflips = tflips;
  ti.vflips = vflips;
  ti.mainloops = mainloops;
  return (&ti);
}

/*--------------------------------------------------------------------------*/

void flip_hook (void (* routine) (char, const int[]))
     /* routine will be called *after* each flip in flip_transform */
{
  hook = routine;
}

/*--------------------------------------------------------------------------*/

void flip_list (int n)
     /* allocates new flip list (with initial size n, but unbounded!);
        deallocates if n <= 0 */
{
  if (nd)
    {
      basic_istaque_dispose (nd);
      nd = NULL;
    }
  if (n > 0)
    nd = basic_istaque_new (n);
}

/*--------------------------------------------------------------------------*/

void flip_list_put (int e)
     /* adds triangle represented by e to the flip list, if non-Delaunay */
{
  if (not is_regular (e))
    {
      char c = trist_color(TrIndex (e));

      if ((c == _WHITE) or (c == _BLUE)) {
	trist_color_set (TrIndex (e), _GREEN);
	add_nd (e);
      }
    }
}

/*--------------------------------------------------------------------------*/

static void crash (void)
{
  print ("\n*** flip_transform CRASHED ***\n flip_dt3: %d + %d non-transf\n",
         reds, greens);
#if REPORT_THE_CRASH
  {
    FILE *ntf;
    ntf = basic_fopen (regular_flip_crash_path, "w");
    ___crash (ntf, reds);
    basic_fclose (ntf);
  }
#else
  print ("COMPILE \"flip.c\" WITH \"%s\" AND RUN AGAIN\n",
         "#define REPORT_THE_CRASH 1");
#endif
  basic_error ("flip_transform CRASHED!!!");
}

/*--------------------------------------------------------------------------*/
/* AUXILIARY ROUTINES FOR MANIPULATING TRIANGLES                            */
/*--------------------------------------------------------------------------*/


static int ch_triangle (int e, int i, int j, int k)
     /* Checks whether e = (i,j,k) */
{
  int i1, j1, k1;

  trist_triangle (e, &i1, &j1, &k1);

  return ((i1 == i) && (j1 == j) && (k1 == k));
}

/*--------------------------------------------------------------------------*/

void pr_triangle (int e)
     /* e is edfacet ijk */
{
  int i, j, k;
  trist_triangle (e, &i, &j, &k);
#ifdef __DEBUG__
  if (trist_test_flag)
    if ((i == j) or (i == k) or (k == j))
      basic_error ("Triangle: Given edge-ring is not triangular.");
#endif
  print("Triangle %d: %d %d %d\n", e, i, j, k);
}

/*--------------------------------------------------------------------------*/

/*
 * Fmerge () -- This function is used by flip_insert to merge two facet
 *              rings.
 */
static void Fmerge (int d, int e)
     /* Merges 2 (nonoverlapping) Facet rings s.t. fnext (d) == e,
        assuming fnext (d) != e. */
{
  Assert ((Org (d) == Org (e)) and (Dest (d) == Dest (e)));
  Fsplice (d, Sym (Fnext (Sym (e))));
}

/*--------------------------------------------------------------------------*/

