/*
 * FILE:  infinite.c
 *   Contains functions dealing with infinite points.
 *
 * Author: Mike Facello
 */

#include "../sos/internal.h"
#include "../sos/primitive.h"
#include "infinite.h"
#include "delcx.h"
#include "dt.h"

/*-------------------------------------------------------------------------*/

Static_Declarations ("lambda_inf", 5, 49, 4);

#include "../sos/primitive.i.c"

/*-------------------------------------------------------------------------*/

/* Local procedures */

static void sos_prim_chs(SoS_primitive_result *result);
static int missing_inf_index(int x, int y, int z);

/*-------------------------------------------------------------------------*/

/* Global variables */

int num_vert;

/*-------------------------------------------------------------------------*/

/* Definitions of coefficients and columns used for positive test  */

short int const4_3inf[4] = {1, -1, 1, -1};

short int const4_2inf[3][3][2] = 
{ { {-1, 2}, {1, 3}, {-1, 3} },
  { {0, 0}, {-1, 3}, {1, 3} },
  { {0, 0}, {0, 0}, {-1, 1} } };

short int const4_1inf[4][2] = 
{ {1, 2}, {-1, 2}, {-1, 1}, {1, 1} } ;

short int const5_3inf[4][2] = 
{ {1, 1}, {1, 1}, {-1, 3}, {1, 3} } ;

short int const5_2inf[3][3][2] =
{ { {1, 2}, {1, 1}, {-1, 1} },
  { {0, 0}, {1, 1}, {-1, 1} },
  { {0, 0}, {0, 0}, {-1, 1} } };

int inf4_count[5] = {0, 0, 0, 0, 0};
int inf5_count[5] = {0, 0, 0, 0, 0};

/*--------------------------------------------------------------------------*/

/*
 * open_infinite (n) -- This function opens the infinite module by
 *   			initializing the number of vertices.  This is
 *                      necessary because the infinite points are
 *                      n+1...n+4.
 */

void open_infinite (int n)
{
/*  print ("Setting num_vert to %d.\n", n); */
  num_vert = n;
}

/*--------------------------------------------------------------------------*/

/*
 * sos_prim_chs (result)
 *          Changes the sign of an sos result.
 */
static void sos_prim_chs (SoS_primitive_result *result)
{
  result->signum = ((result->signum > 0) ? -1 : 1);
  lia_chs(result->lia_pointer);
}

/*-------------------------------------------------------------------------*/

/*
 * num_infinite4(w, x, y, z)
 *       Returns the number of infinite points in w, x, y, z
 */
int num_infinite4(int w, int x, int y, int z)
{
  return (is_infinite(w) + is_infinite(x) + is_infinite(y) + is_infinite(z));
}

/*-------------------------------------------------------------------------*/

/* 
 * num_infinite5(v, w, x, y, z) 
 * Returns the number of infinite points in v, w, x, y, z 
 */ 
int num_infinite5(int v, int w, int x, int y, int z) 
{ 
  return (is_infinite(v) + is_infinite(w) + is_infinite(x) +
	  is_infinite(y) + is_infinite(z)); 
}

/*-------------------------------------------------------------------------*/

/*
 * missing_inf_index (x, y, z) --
 *        Assuming that x, y, and z are the indices of infinite points
 *        in sorted order, returns 0, 1, 2 or 3 indicating which index is
 *        missing.
 */
static int missing_inf_index(int x, int y, int z)
{
  if (x != 1)
    return(0);
  if (y != 2)
    return(1);
  if (z != 3)
    return(2);
  return(3);
}

/*-------------------------------------------------------------------------*/

/*
 * sos_lambda4_inf(i, j, k, l)
 *            Returns the significant term of the Lambda4 epsilon
 *            determinant.  Assumes indices are in proper range,
 *            pairwise different, and sorted.
 *
 *            BIG NOTE:  Due to the fact that the infinite vertices used
 *            to be 1..4, and now are n+1..n+4, and thus are at the end
 *            of the sorted list rather than the beginning, some signs
 *            need to be changed (marked by **** below).
 */

SoS_primitive_result *sos_lambda4_inf (int i, int j, int k, int l)
{
  SoS_primitive_result *temp;
  int num_inf = num_infinite4(i, j, k, l);

/*  print ("sos_lambda4_inf w/ params (%d, %d, %d, %d)\n", i, j, k, l); */
  
  inf4_count[num_inf] ++;

/*  print ("num_inf = %d\n", num_inf); */
  
  switch (num_inf) {
  case 0:
    return(sos_lambda4 (i, j, k, l));
    break;
  case 1:
    l -= num_vert;
    /* BEFORE: temp = sos_lambda3_col(j, k, l, col4_1inf(i), 3); */
    temp = sos_lambda3_col(i, j, k, col4_1inf(l), 3);
    if (coef4_1inf(l) != -1)                              /****/
      sos_prim_chs(temp);
    return(temp);
  case 2:
    k -= num_vert; l -= num_vert;
    /* BEFORE: temp = sos_lambda2_col( k, l, col4_2inf(i, j)); */
    temp = sos_lambda2_col( i, j, col4_2inf(k, l));
    if (coef4_2inf(k, l) == -1)
      sos_prim_chs(temp);
    return(temp);
  case 3:
    j -= num_vert; k -= num_vert; l -= num_vert; 
    /* BEFORE: result.signum = - sign4_3inf(i, j, k); */
    result.signum = - sign4_3inf(j, k, l);		  /****/
    result.lia_pointer = lia_const(result.signum);
    result.depth = 0;
    return (&result);
  case 4:
    result.signum = -1;
    result.lia_pointer = lia_const(result.signum);
    result.depth = 0;
    return (&result);	/* In sorted order, infinite points are 
				   in negative orientation */
  }
}

/*-------------------------------------------------------------------------*/

/*
 * sos_lambda5_inf(i, j, k, l, m)
 *            Returns the significant term of the Lambda4 epsilon
 *            determinant.  Assumes indices are in proper range,
 *            pairwise different, and sorted.
 *
 *            BIG NOTE:  Due to the fact that the infinite vertices used
 *            to be 1..4, and now are n+1..n+4, the infinite points are at 
 *            the end of the sorted list rather than the beginning.
 *            However, in the case of a 5x5 determinant, this does not
 *            matter.  The number of swaps necessary to get it into the
 *            original order is always even.
 */

SoS_primitive_result *sos_lambda5_inf (int i, int j, int k, int l, int m)
{
  SoS_primitive_result *temp;
  int num_inf = num_infinite5(i, j, k, l, m);

  inf5_count[num_inf] ++;

  switch (num_inf) {
  case 0:
    return(sos_lambda5 (i, j, k, l, m));
    break;
  case 1:
    /* BEFORE: temp = sos_lambda4(j, k, l, m); */
    temp = sos_lambda4(i, j, k, l);
    sos_prim_chs(temp);
    return(temp);
  case 2:
    l -= num_vert; m -= num_vert;
    /* BEFORE: temp = sos_lambda3_col( k, l, m, col5_2inf(i, j), 3); */
    temp = sos_lambda3_col( i, j, k, col5_2inf(l, m), 3);
    if (coef5_2inf(l, m) == -1)
      sos_prim_chs(temp); 
   return(temp);
  case 3:
    k -= num_vert; l -= num_vert; m -= num_vert;
    /* BEFORE: temp = sos_lambda2_col( l, m, col5_3inf(i, j, k)); */
    temp = sos_lambda2_col( i, j, col5_3inf(k, l, m));
    if (coef5_3inf(k, l, m) == -1)
      sos_prim_chs(temp);
    return (temp);
  case 4:
    result.signum = -1;
    result.lia_pointer = lia_const(result.signum);
    result.depth = 0;
    return (&result);		/* In sorted order, this determinant is
				   always negative */
  }
}



/*--------------------------------------------------------------------------*/

int sos_in_sphere_inf (int o, int i, int j, int k, int p)
     /* SoS predicate in_sphere, aka in_ball3.
        Assumes indices in proper range and pairwise different. */
{
  int result, s, d1, d2;
#ifdef __SOS_DEBUG__
  print ("sos_in_sphere (%d,%d,%d,%d,%d)\n", o, i, j, k, p);
#endif
  (void) basic_isort4 (&o, &i, &j, &k);
  d1 = sos_lambda4_inf (o, i, j, k) -> signum;
  s = basic_isort5p (&o, &i, &j, &k, &p);
  d2 = sos_lambda5_inf (o, i, j, k, p) -> signum;
  result = If (Odd (s), (d1 != d2), (d1 == d2));
  /* Note: This is wrong in \cite{edels:sos}. */
#ifdef __SOS_DEBUG__
  if (not ((0 < o) and (o < i) and (i < j) and (j < k) and (k < p)
	   and (p <= sos_common.n)))
    basic_error ("sos_in_sphere: arguments were wrong");
  print ("sos_in_sphere (%d,%d,%d,%d,%d) result: %d (%d %d %d)\n",
	 o, i, j, k, p, result, s, d1, d2);
#endif
  return (result);
}

/*--------------------------------------------------------------------------*/

int sos_in_sphere_inf_p (int o, int i, int j, int k, int p)
     /* SoS predicate in_sphere_p, aka in_ball3_p.
        Assumes indices in proper range and pairwise different.
        Identical to sos_in_sphere() but with the assumption that
        sos_positive3 (o, i, j, k) == TRUE. */
{
  int result, s = 0, d2;
#ifdef __SOS_DEBUG__
  print ("sos_in_sphere_p (%d,%d,%d,%d,%d,%c) ", o, i, j, k, p);
  {
    int d1;
    s = basic_isort4 (&o, &i, &j, &k);
    d1 = sos_lambda4_inf (o, i, j, k) -> signum;
    if (Odd (s))
      d1 = -d1;
    if (d1 != 1)
      basic_error ("sos_in_sphere_p: orientation not positve");
  }
#endif
  s += basic_isort4 (&o, &i, &j, &k);
  s += basic_isort5p (&o, &i, &j, &k, &p);
  d2 = sos_lambda5_inf (o, i, j, k, p) -> signum;
  result = If (Odd (s), (1 == d2), (1 != d2));
#ifdef __SOS_DEBUG__
  if (not ((0 < o) and (o < i) and (i < j) and (j < k) and (k < p) and
	   (p <= sos_common.n)))
    basic_error ("sos_in_sphere_p: arguments were wrong");
  print ("sos_in_sphere_p (%d,%d,%d,%d,%d) result: %d (%d %d)\n",
	 result, s, d2);
#endif
  return (result);
}

/*--------------------------------------------------------------------------*/

int sos_positive3_inf (int h, int i, int j, int k)
     /* SoS predicate positive3.
        Assumes indices in proper range and pairwise different. */
{
  int result, s, d;
#ifdef __SOS_DEBUG__
  print ("sos_positive3 (%d,%d,%d,%d) ", h, i, j, k); 
#endif
  s = basic_isort4 (&h, &i, &j, &k);
  d = sos_lambda4_inf (h, i, j, k) -> signum;

  result = If (Odd (s), (d == -1), (d == 1));
#ifdef __SOS_DEBUG__
  if (not ((0 < h) and (h < i) and (i < j) and (j < k)
	   and (k <= sos_common.n)))
    basic_error ("sos_positive3: arguments were wrong");
  print  ("sos_positive3 (%d,%d,%d,%d) result: %d (%d %d)\n",
	  h, i, j, k, result, s, d);
#endif
  return (result);
}

/*-------------------------------------------------------------------------*/

/* 
 * infinite_summary (file) -- 
 *            Prints a summary of the performance of the infinite point
 *            procedures.
 */
void infinite_summary (FILE *file)
{
  int i;

  fprint (file, "* SoS infinite tests - lambda4\n");
  for (i = 0; i <= 4; i++)
    fprint (file, "%12d . %d infinite points\n", inf4_count[i], i);

  fprint (file, "* SoS infinite tests - lambda5\n");
  for (i = 0; i <= 4; i++)
    fprint (file, "%12d . %d infinite points\n", inf5_count[i], i);
}
