/*
 * primflip.c -- Primitive flip routines for Delaunay triangulations.
 *
 * author -- Mike Facello (based on Ernst Mucke's original code)
 *
 * date -- 2/17/96
 */

/*--------------------------------------------------------------------------*/

#include "primflip.h"

static void delete_triangle (int ef);
static void load (int xyz, int x, int y, int z, int o, int p, 
		  int field[/*10*/]);
static void load2 (int pox, int x, int y, int z, int o, int p, 
		   int field[/*10*/]);
static void load3 (int e, /* i, j, k, l, m, */ int field[/*10*/]);

static void diagonalize (int xyz, int field[/*10*/], 
			 int x, int y, int z, int o, int p);
static int diagonalize_test (int xpo, int x, int y, int z, int p, int o);
static void cut (int field[/*10*/], int x, int y, int z, int o, int p);

static void Fmerge(int d, int e);

int print_triangle(int ef);

static int num_inserts = 0, num_tflips = 0, num_eflips = 0, num_vflips = 0;

#define FLIP_TEST FALSE

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 * delete_triangle () -- This function removes triangle e from the
 *                           triangulation without deleting the triangle
 *                           record.
 */
static void delete_triangle (int ef)
{
  register int ef2 = Enext (ef), ef3 = Enext2 (ef);

  Fsplice (Sym (Fnext (Sym (ef))),  ef);
  Fsplice (Sym (Fnext (Sym (ef2))), ef2);
  Fsplice (Sym (Fnext (Sym (ef3))), ef3);
}

/*--------------------------------------------------------------------------*/

/*
 * prim_flip_insert (ijk, v) -- Insert vertex v inside the tetrahedra 
 *                         containing triangle ijk.  The four new tetrahedra
 *                         are created, and the search data structure is 
 *                         updated.
 */
void prim_flip_insert (int ijk, int v, int field[/*10*/])
{
  int i, j, k, l;
  int jil, kjl, ikl, ijv, jkv, ikv, ilv, jlv, klv;
  int t;                                /* The root of the search tree */

  num_inserts++;

  /* Find the other three faces of the tetrahedra */

  jil = Sym (Fnext (ijk));
  kjl = Sym (Fnext (Enext (ijk)));
  ikl = Sym (Fnext (Enext2 (ijk)));

  /* Find the vertex numbers */

  trist_tetra(ijk, &i, &j, &k, &l);

#ifdef SDEBUG
  print ("Testing that point is inside tetrahedron\n");
  print ("(%d, %d, %d, %d) = %d\n", v, i, j, k, sos_positive3_inf (v,i,j,k));
  print ("(%d, %d, %d, %d) = %d\n", v, j, i, l, sos_positive3_inf (v,j,i,l));
  print ("(%d, %d, %d, %d) = %d\n", v, k, j, l, sos_positive3_inf (v,k,j,l));
  print ("(%d, %d, %d, %d) = %d\n", v, i, k, l, sos_positive3_inf (v,i,k,l));

   Assert_always (    sos_positive3_inf (v,i,j,k)
		 and sos_positive3_inf (v,j,i,l)
		 and sos_positive3_inf (v,k,j,l)
		 and sos_positive3_inf (v,i,k,l));
/*
  Assert_always (    sos_positive3 (v,i,j,k)
		 and sos_positive3 (v,j,i,l)
		 and sos_positive3 (v,k,j,l)
		 and sos_positive3 (v,i,k,l));
*/

  printf("Inserting vertex %d in tetrahedra (%d, %d, %d, %d).\n",
	  v, i, j, k, l);
#endif

  /* Create the 6 new triangles that are needed */

  ijv = trist_make(i, j, v);
  jkv = trist_make(j, k, v);
  ikv = trist_make(i, k, v);
  ilv = trist_make(i, l, v);
  jlv = trist_make(j, l, v);
  klv = trist_make(k, l, v);

  t = TrIndex (jlv);

  /* Update the face rings for the tetrahedra faces */

  Fmerge(ijk, ijv);
  Fmerge(Enext(ijk), jkv);
  Fmerge(Enext2(ijk), Sym(ikv));
  Fmerge(Enext(jil), ilv);
  Fmerge(Enext(ikl), klv);
  Fmerge(Enext(kjl), jlv);

  /* Create the rings containing only the new faces */

  /* vij -> vil -> vik */

  Fmerge(Enext2(ijv), Enext2(ilv));
  Fmerge(Enext2(ilv), Enext2(ikv));

  /* vjk -> vjl -> vji */

  Fmerge(Enext2(jkv), Enext2(jlv));
  Fmerge(Enext2(jlv), Sym(Enext(ijv)));

  /* kvi -> kvj -> kvl */

  Fmerge(Enext(ikv), Enext(jkv));
  Fmerge(Enext(jkv), Enext2(Sym(Enext(klv))));

  /* lvi -> lvk -> lvj */

  Fmerge(Enext(ilv), Enext(klv));
  Fmerge(Enext(klv), Enext(jlv));

  /* Now, test all of the tetrahedra that were created */

#ifdef SDEBUG
  dt_test_triangle(ijv, i, j, v);
  dt_test_triangle(ikv, i, k, v);
  dt_test_triangle(ilv, i, l, v);
  dt_test_triangle(jkv, j, k, v);
  dt_test_triangle(jlv, j, l, v);
  dt_test_triangle(klv, k, l, v);
  dt_test_open_tetra(ijv, i, j, k, v);
  dt_test_open_tetra(Sym(ijv), j, i, l, v);
  dt_test_open_tetra(Sym(jkv), k, j, l, v);
  dt_test_open_tetra(ikv, i, k, l, v);
#endif

  /* We also need to update the search data structure */

  field[0] = ijv; field[1] = ikv; field[2] = jkv;
  field[3] = ilv; field[4] = jlv; field[5] = klv;
  field[6] = jil; field[7] = kjl; field[8] = ikl;
  field[9] = ijk;

/*  search_update_insert (v, ijk, field, i, j, k, l); */

  /* Finally, put all the triangles onto the flip list. */
/*
  flip_list_put(ijk);
  flip_list_put(jil);
  flip_list_put(kjl);
  flip_list_put(ikl);
  set_blue (TrIndex(jlv));
  set_blue (TrIndex(ilv));
  set_blue (TrIndex(klv));
  set_blue (TrIndex(ikv));
  set_blue (TrIndex(ijv));
  set_blue (TrIndex(jkv));
*/
}


/*--------------------------------------------------------------------------*/

/*
 * prim_flip_triangle (xyz) -- attempts to flip the triangle xyz.  Assumes that
 *                        xyz is a link facet, and that p_i, the inserted
 *                        point is on the positive side of xyz.  This means
 *                        that o in the code below is p_i.
 */
void prim_flip_triangle (int xyz, int field[/*10*/])
     /* assumes is_regular (xyz) == FALSE */
{
  int x, y, z, o, p, xyo = Fnext (xyz), j;

  num_tflips++;

  trist_triangle (xyz, &x, &y, &z);
  o = Dest (Enext (xyo));
  p = Dest (Enext (Fnext (Sym (xyz))));

  Assert_if (flip_test_flag,
             (    sos_positive3_inf (o, x, y, z)
              and (not sos_positive3_inf (p, x, y, z))
              and (not sos_in_sphere_inf_p (o, x, y, z, p))
              and dt_test_triangle (xyo, x, y, o)
              and dt_test_triangle (Fnext (Sym (xyz)), y, x, p)));

  load (xyz, x, y, z, o, p, field);
  delete_triangle (xyz);
  diagonalize (xyz, field, x, y, z, o, p);
}


/*--------------------------------------------------------------------------*/

/*
 * prim_flip_edge (pox, field) --    Flips the edge op.  Assumes that
 *                        pox is a link facet, and that p_i, the inserted
 *                        point is on the positive side of pox.  This means
 *                        that y in the code below is p_i.
 */
void prim_flip_edge (int pox, int field[/*10*/])
{
  int x, y, z, o, p, j;
  int poy, poz;

  num_eflips++;

  trist_triangle (pox, &p,  &o,  &x);
  poy = Fnext (pox);
  poz = Fnext (poy);
  y = Dest (Enext (poy));
  z = Dest (Enext (poz));

  Assert_if (flip_test_flag,
             (    sos_positive3_inf (o, x, y, z)
              and (not sos_positive3_inf (p, x, y, z))
              and (not sos_positive3_inf (p, y, z, o))
              and (not sos_positive3_inf (p, z, x, o))
              and (not sos_positive3_inf (p, x, y, o))));
  Assert_if (flip_test_flag,
             sos_in_sphere_inf_p (o, x, y, z, p));

  /* flip edge to triangle;
     assuming sos_positive3 (o, x, y, z) and sos_positive3 (p, y, x, z) */

  load2 (pox, x, y, z, o, p, field);
  delete_triangle (pox);         /* Each of these triangles is now */
  delete_triangle (poy);         /* a search triangle.             */
  delete_triangle (poz);
  cut (field, x, y, z, o, p);

#ifdef NOWAY
  search_update_edge (pox, poy, poz, field, x, y, z, o, p);

  /* Only link facets are checked */
  check_triangle (Sym(field[2]));        /* xzo */
  check_triangle (Sym(field[4]));        /* zxp */
  set_blue (TrIndex(field[0]));
  set_blue (TrIndex(field[1]));
  set_blue (TrIndex(field[3]));
  set_blue (TrIndex(field[5]));

  if (hook)
    {
      int xyz = Sym (Fnext (Sym (xyo)));
      Assert (dt_test_triangle (xyz, x, y, z));
      hook (FLIP_EDGE, x, y, z, o, p, xyz);                
    }
  return (TRUE);
#endif

}

/*--------------------------------------------------------------------------*/

/*
 * prim_flip_vertex (ivj, field[]) --
 *        This procedure will attempt to flip the vertex that is the
 *        destination of triangle ef.
 */
void prim_flip_vertex (int ivj, int field[/*10*/])
{
  int i;

#ifdef VDEBUG
  print ("&&&& Performing vertex flip with vertex %d\n", Dest(ivj));
#endif

  num_vflips++;

  load3 (ivj, field);
  upfor (i, 0, 5)
    delete_triangle (field[i]);

  /* field[8] is the only new link facet
  flip_list_put (field[8]);

  add_redundant_vertex(Dest(e));

  search_update_vertex (field);
  
  return (TRUE);
*/

}


/*--------------------------------------------------------------------------*/

/*
 * load () -- These are used to compute the triangles involved
 *                         in the flips and return them in an array.  load
 * 			   is used by flip_triangle.
 */

static void load (int xyz, int x, int y, int z, int o, int p, 
		  int field[/*10*/])
     /* used to return:  
	          field[0] = xyo,  field[1] = yzo,  field[2] = zxo
                  field[3] = yxp,  field[4] = xzp,  field[5] = zyp */
     /* Now returns:
	   Deleted:
 	       field[0] = xyz,   
           Added:
               field[1] = pxo,   field[2] = pyo,   field[3] = pzo,   
           Link:
               field[4] = yxo,   field[5] = xzo,   field[6] = zyo,   
	       field[7] = xyp,   field[8] = yzp,   field[9] = zxp
                                                                   */
     /* Conversion:
	old field[0] = Sym(field[4]);
	old field[1] = Sym(field[6]);
	old field[2] = Sym(field[5]);
	old field[3] = Sym(field[7]);
	old field[4] = Sym(field[9]);
	old field[5] = Sym(field[8]);
	*/
{
  field[0] = xyz;                       /* xyz */
  Assert (dt_test_triangle (field[0], x, y, z));
  field[4] = Sym(Fnext(xyz));           /* yxo */
  Assert (dt_test_triangle (field[4], y, x, o));
  field[5] = Enext (Sym (Fnext (Enext (field[4]))));            /* xzo */
  Assert (dt_test_triangle (field[5], x, z, o));
  field[6] = Enext (Sym (Fnext (Enext (field[5]))));            /* zyo */
  Assert (dt_test_triangle (field[6], z, y, o));
  field[7] = Sym(Fnext(Sym(xyz)));      /* xyp */
  Assert (dt_test_triangle (field[7], x, y, p));
  field[8] = Enext (Sym (Fnext (Enext (field[7]))));            /* yzp */
  Assert (dt_test_triangle (field[8], y, z, p));
  field[9] = Enext (Sym (Fnext (Enext (field[8]))));            /* zxp */
  Assert (dt_test_triangle (field[9], z, x, p));
}

/*--------------------------------------------------------------------------*/

static void load2 (int pox, int x, int y, int z, int o, int p, 
		   int field[/*10*/])
     /* used to return:  
	          field[0] = xyo,  field[1] = yzo,  field[2] = zxo
                  field[3] = yxp,  field[4] = xzp,  field[5] = zyp */
     /* Now returns:
           Deleted:
	       field[0] = pox,   field[1] = poy,   field[2] = poz,
           Added:
	       field[3] = xyz,   
           Link:
               field[4] = oyx,   field[5] = ypx,   field[6] = ozy,   
	       field[7] = zpy,   field[8] = oxz,   field[9] = xpz
                                                                     */
     /* Conversion:
	old field[0] = Sym(Enext(field[4]));
	old field[1] = Sym(Enext(field[6]));
	old field[2] = Sym(Enext(field[8]));
	old field[3] = Sym(Enext2(field[5]));
	old field[4] = Sym(Enext2(field[9]));
	old field[5] = Sym(Enext2(field[7]));
	*/
{
  int h;

  field[0] = pox;                       /* pox */
  Assert (dt_test_triangle (field[0], p, o, x));
  field[1] = Fnext(pox);                /* poy */
  Assert (dt_test_triangle (field[1], p, o, y));
  field[2] = Fnext(field[1]);           /* poz */
  Assert (dt_test_triangle (field[2], p, o, z));
  field[4] = Enext (Sym (Fnext (Enext (field[0]))));        /* oyx */
/*  field[4] = Turn(field[0]);        /* oyx */
  Assert (dt_test_triangle (field[4], o, y, x));
  field[5] = Enext (Sym (Fnext (Enext (field[4]))));                   /* ypx */
/*  field[5] = Turn(field[4]);                   /* ypx */
  Assert (dt_test_triangle (field[5], y, p, x));
  field[6] = Enext (Sym (Fnext (Enext (field[1]))));        /* ozy */
/*  field[6] = Turn(field[1]);        /* ozy */
  Assert (dt_test_triangle (field[6], o, z, y));
  field[7] = Enext (Sym (Fnext (Enext (field[6]))));                   /* zpy */
/*  field[7] = Turn(field[6]);                   /* zpy */
  Assert (dt_test_triangle (field[7], z, p, y));
  field[8] = Enext (Sym (Fnext (Enext (field[2]))));        /* oxz */
/*  field[8] = Turn(field[2]);        /* oxz */
  Assert (dt_test_triangle (field[8], o, x, z));
  field[9] = Enext (Sym (Fnext (Enext (field[8]))));                   /* xpz */
/*  field[9] = Turn(field[9]);                   /* xpz */
  Assert (dt_test_triangle (field[9], x, p, z));
}

/*--------------------------------------------------------------------------*/

/*
 * load3 (ivj, field) --
 *           This procedure loads the array field with triangles needed
 *           for a vertex flip.  NOTE:  The vertex being deleted is j!!!
 */
static void load3 (int ivj, int field[/*10*/])
     /* Returns:
	   field[0] = ivj,   field[1] = ivk,   field[2] = ivl,
	   field[3] = vjk,   field[4] = vkl,   field[5] = vlj,
	   field[6] = ijk,   field[7] = ikl,   field[8] = ilj,
	   field[9] = kjl                                          */

{
  /* If e = ivj, then field is loaded as follows:  */

  field[0] = ivj;				/* ivj */
  field[1] = Fnext(ivj);			/* ivk */
  field[2] = Fnext(field[1]);			/* ivl */
  field[3] = Fnext(Enext(ivj));			/* vjk */
  field[4] = Fnext(Enext(field[1]));		/* vkl */
  field[5] = Fnext(Enext(field[2]));		/* vlj */
  field[6] = Sym(Fnext(Enext2(ivj)));		/* ijk */
  field[7] = Sym(Fnext(Enext2(field[1])));	/* ikl */
  field[8] = Sym(Fnext(Enext2(field[2])));	/* ilj */
  field[9] = Sym(Fnext(Enext(field[3])));	/* kjl */

#ifdef DEBUG
  print ("Checking that m is actually p_i\n");
#endif

/*
 *  This can be used if the vertices are passed in as parameters
 *
#ifdef VDEBUG
  dt_test_triangle(field[0], i, v, j);
  dt_test_triangle(field[1], i, v, k);
  dt_test_triangle(field[2], i, v, l);
  dt_test_triangle(field[3], v, j, k);
  dt_test_triangle(field[4], v, k, l);
  dt_test_triangle(field[5], v, l, j);
  dt_test_triangle(field[6], i, j, k);
  dt_test_triangle(field[7], i, k, l);
  dt_test_triangle(field[8], i, l, j);
  dt_test_triangle(field[9], l, j, l);
#endif
*/
}


/*--------------------------------------------------------------------------*/

/*
 * diagonalize () and cut ()  -- This two functions are refinements of
 *        flip_triangle and flip_edge respectively.  They perform the
 *        trist manipulations necessary to update the triangle
 *        structure.
 */

static void diagonalize (int xyz, int field[/*10*/], 
			 int x, int y, int z, int o, int p)
     /* assumes that we have a double tetrahedron;
         creates 3 new triangles, i.e, 3*3 edfacets   */
{
  int xpo, pyo, pzo, oxz, zpx, xpz, zox, yoz, pyz;
  xpo = trist_make (x, p, o);
  field[1] = Sym(xpo);
  field[2] = pyo = trist_make (p, y, o);
  field[3] = pzo = trist_make (p, z, o);
  
/*
  oxz = Sym (Enext (field[2]));
  zpx = Enext (field[4]);
  xpz = Sym (Enext (zpx));
  zox = Enext2 (oxz);
  yoz = Enext (Sym (field[1]));
  pyz = Sym (Enext (field[5]));
*/

  oxz = Enext2 (field[5]);
  Assert (dt_test_triangle(oxz, o, x, z));
  xpz = Enext (field[9]);
  Assert (dt_test_triangle(xpz, x, p, z));
  zpx = Sym (Enext (xpz));
  Assert (dt_test_triangle(zpx, z, p, x));
  zox = Enext2 (oxz);
  Assert (dt_test_triangle(zox, z, o, x));
  yoz = Enext (field[6]);
  Assert (dt_test_triangle(yoz, y, o, z));
  pyz = Enext2 (field[8]);
  Assert (dt_test_triangle(pyz, p, y, z));

  Fsplice (xpz, xpo);
  Fsplice (oxz, Enext2 (xpo));
  Fsplice (pyz, pyo);
  Fsplice (yoz, Enext (pyo));
  Fsplice (Sym (zpx), pzo);
  Fsplice (zox, Enext (pzo));
  Fsplice (Enext (xpo), Enext (Sym (pyo)));
  Fsplice (Enext (Sym (pyo)), Enext (Sym (pzo)));
  Assert (diagonalize_test (xpo, x, y, z, p, o));
}

/*--------------------------------------------------------------------------*/

#if __DEBUG__
static int diagonalize_test (int xpo, int x, int y, int z, int p, int o)
{
  int one, two, three;
  one   = Enext (xpo);
  two   = Fnext (one);
  three = Fnext (two);
  Assert_always (    (one == Fnext (three))
                 and dt_test_triangle (one,   p, o, x)
                 and dt_test_triangle (two,   p, o, y)
                 and dt_test_triangle (three, p, o, z)
                 and dt_test_open_tetra (one,   p, o, z, x)
                 and dt_test_open_tetra (two,   p, o, x, y)
                 and dt_test_open_tetra (three, p, o, y, z));
  return (TRUE);
}
#endif

/*--------------------------------------------------------------------------*/

static void cut (int field[/*10*/], int x, int y, int z, int o, int p)
     /* assumes that we have a double tetrahedron;
	creates 1 new triangle, i.e, 1*3 edfacets + syms !! (every else too)*/
{
  int xyz;
  field[3] = xyz = trist_make (x, y, z);

/*  
  Fsplice (Sym (Fnext (Sym (field[0]))), xyz);                    xyp!
  Fsplice (Sym (Fnext (Sym (field[1]))), Enext (xyz));            yzp!
  Fsplice (Sym (Fnext (Sym (field[2]))), Enext (Enext (xyz)));    zxp!
*/

  Fsplice (Enext2 (field[5]), xyz);                    /* xyp */
  Fsplice (Enext2 (field[7]), Enext (xyz));            /* yzp */
  Fsplice (Enext2 (field[9]), Enext (Enext (xyz)));    /* zxp */

  Assert_if (flip_test_flag,
             (    dt_test_open_tetra (Fnext (xyz), x, y, z, o)
              and dt_test_open_tetra (Fnext (Sym (xyz)), y, x, z, p)
	      and dt_test_triangle (Fnext (Fnext (Enext (field[4]))), 
				    y, x, p)));
}

/*--------------------------------------------------------------------------*/

/*
 * Fmerge () -- This function is used by flip_insert to merge two facet
 *              rings.
 */
static void Fmerge (int d, int e)
     /* Merges 2 (nonoverlapping) Facet rings s.t. fnext (d) == e,
        assuming fnext (d) != e. */
{
  Assert ((Org (d) == Org (e)) and (Dest (d) == Dest (e)));
  Fsplice (d, Sym (Fnext (Sym (e))));
}

/*--------------------------------------------------------------------------*/

int print_triangle(int ef)
{
  int x, y, z;

  trist_triangle (ef, &x, &y, &z);
  print ("Triangle %d = (%d, %d, %d)\n", ef, x, y, z);
}

/*--------------------------------------------------------------------------*/

/*
 * prim_flip_info () -- This function returns information about the number
 *                      of primitive flip operations performed.
 */
Prim_flip_info *prim_flip_info (void)
{
  static Prim_flip_info pinfo;

  pinfo.inserts = num_inserts;
  pinfo.tflips  = num_tflips;
  pinfo.eflips  = num_eflips;
  pinfo.vflips  = num_vflips;

  return &pinfo;
}
