/*
 * search.c -- Routines for point location
 *
 * author -- Michael Facello
 *
 * Modifications
 *   	7/5/91  -- Added definition of Init_Outer_Triangle and LoadPoint
 *      7/12/91 -- Added inside
 *      4/18/92 -- Modifications for combined search/trist structure
 *      6/2/94  -- Modifications for using standard trist structure
 */
#include "../lia/base.h"
#include "delcx.h"
#include "dt.h"

/* Global variables */

int in_tests = 0;
int ivertex;

int stype = CASCADE_SEARCH;

/* Local variables */

long int tot_vis = 0;	       /* total # of nodes visited during searches */
int not_inserted = 0;          /* number of vertices not inserted */
long vis_not_inserted = 0;     /* total depth of search before not inserted */

/* Function variables for generic search */

void (* search_init)(int vertex[], int n) = NULL;
int  (* search_make)(int i, int j, int k) = NULL;
int  (* search_current_triangle)(int t) = NULL;
int  (* search_locate_vertex)(int v) = NULL;
int  (* search_locate_redundant)(int v) = NULL;
void (* search_update_insert)(int v, int ijk, const int field[], 
			      int i, int j, int k, int l) = NULL;
#ifdef OLDSTUFF
void (* search_update_triangle)(int xyz, const int field[/*5*/],
				int x, int y, int z, int o, int p) = NULL;
void (* search_update_edge)(int pox, int poy, int poz, const int field[/*6*/], 
			    int x, int y, int z, int o, int p) = NULL;
void (* search_update_vertex)(const int field[/*9*/]) = NULL;
#endif
void (* search_update_triangle)(int x, int y, int z, int o, int p,
				const int field[/*5*/]) = NULL;
void (* search_update_edge)(int x, int y, int z, int o, int p,
			    const int field[/*6*/]) = NULL;
void (* search_update_vertex)(const int field[/*9*/]) = NULL;
void (* search_close)(void) = NULL;

/* Local functions */

/* Functions for implementing no search. */

static void nsearch_init_func (int vertex[], int n);
static int  nsearch_search_func (int v);
static void nsearch_update_insert   (int v, int ijk, const int field[], 
				     int i, int j, int k, int l);
static void nsearch_update_triangle (int x, int y, int z, int o, int p,
				     const int field[/*10*/]);
static void nsearch_update_edge     (int x, int y, int z, int o, int p,
				     const int field[/*10*/]);
static void nsearch_update_vertex   (const int field[/*10*/]);
static void nsearch_close (void);



/* Auxiliary functions */

static void count_inf_edges(int *ext_e);

static int edge_index (int e, int dope[]);
static int *new_edge_dope (int *max_index);
static void kill_edge_dope (int *dope);

static int outer_tetra(int ijk), outer_triangle(int tri), outer_edge(int e);
static void print_search_helper(int t, int level);

static void single_test(int i, int j, int k, int l);
static void  test_infinite(void);


/*------------------------------------------------------------------------*/

/*
 * search_open () -- This function initializes the search module by 
 *                   setting the search function variables.
 */

void search_open (int search_type)
{
  stype = search_type;

  switch (search_type)
    {
    case TRIDAG_SEARCH:
      trist_data_size (sizeof (Dag_record));
      trist_data_zero ();
      search_init = tridag_init;
      search_make = tridag_make;
      search_current_triangle = is_current_triangle;
      search_locate_vertex = tridag_locate_vertex;
      search_locate_redundant = tridag_locate_redundant;
      search_update_insert = tridag_update_insert;
      search_update_triangle = tridag_update_triangle;
      search_update_edge = tridag_update_edge;
      search_update_vertex = tridag_update_vertex;
      search_close = tridag_kill;
      break;

    case CASCADE_SEARCH:
      trist_data_size (sizeof (Cascade_record));
      trist_data_zero ();
      search_init = cascade_init;
      search_make = cascade_make;
      search_current_triangle = trist_deleted;
      search_locate_vertex = cascade_locate_vertex;
      search_locate_redundant = cascade_locate_redundant;
      search_update_insert = cascade_update_insert;
      search_update_triangle = cascade_update_triangle;
      search_update_edge = cascade_update_edge;
      search_update_vertex = cascade_update_vertex;
      search_close = cascade_close;
      break;

    case NO_SEARCH:
      search_init = nsearch_init_func;
      search_make = trist_make;
      search_current_triangle = trist_deleted;
      search_locate_vertex = nsearch_search_func;
      search_locate_redundant = nsearch_search_func;
      search_update_insert = nsearch_update_insert;
      search_update_triangle = nsearch_update_triangle;
      search_update_edge = nsearch_update_edge;
      search_update_vertex = nsearch_update_vertex;
      search_close = nsearch_close;
      break;

    default:
      basic_error ("Unknown search type\n");
    }
}



/*-----------------------------------------------------------------------*/

/*
 * search_insert (v) -- Inserts the vertex v into the triangulation by
 *              searching for the tetrahedra that it lies in, and then
 *              inserting the point in this tetrahedra.
 */
int search_insert(int v)
{
  int t;
  int flips = 0;	 /* Number of flips performed */

#ifdef __DEBUG__
  printf ("Inserting Vertex %d\n", v);
#endif

  t = search_locate_vertex(v);

  if (t > 0) {				/* t <= 0 means that v is redundant */
    flip_insert(t, v);

#ifdef SDEBUG2
    printf ("dag before flipping\n");
    PrintSearchDag();
#endif

    flips = flip_transform();

#ifdef SDEBUG2
    printf ("dag after flipping\n");
    PrintSearchDag();
#endif
  }

  return (flips);
}

/*-----------------------------------------------------------------------*/

/*
 * in_tetra () -- This function tests if v is inside the tetrahedra 
 *                determined by i, j, k, l, assumed to be in positive
 *                orientation.
 */
int search_in_tetra (int i, int j, int k, int l, int v)
{
  return (    sos_positive3_inf (v,i,j,k)
	  and sos_positive3_inf (v,j,i,l)
	  and sos_positive3_inf (v,k,j,l)
	  and sos_positive3_inf (v,i,k,l));
}

/*-----------------------------------------------------------------------*/

/*
 * search_pos_orientation (v, e) -- This procedure will return 1 or 0 
 *                           depending on whether v has positive orientation 
 *                           with respect to e.
 *
 *                           Note: this procedure used to be for a
 *                           triangle index, not an edfacet.
 */
int search_pos_orientation (int v, int e)
{
  int x, y, z;
  int temp;

  trist_triangle (e, &x, &y, &z);

  /* update stats */
  sos_pos3[ num_infinite4(v, x, y, z) ][SEARCH] ++;

  temp = sos_positive3_inf (v, x, y, z);

#ifdef CDEBUG
  print ("            Orientation of %d with (%d, %d, %d) is %d\n",
	 v, x, y, z, temp);
#endif

  return (temp);
}

/*-------------------------------------------------------------------------*/

/*
 * search_is_infinite_triangle (e) -- returns TRUE if t contains one of the
 *		infinite points
 */
int search_is_infinite_triangle (int e)
{
  int x, y, z;

  trist_triangle (e, &x, &y, &z);
  return (is_infinite (x) or is_infinite (y) or is_infinite (z));
}

/*-------------------------------------------------------------------------*/

/*
 * kill_infinite_triangles () -- This function will delete all infinite
 *             triangles, and relabel the new convex hull triangles.
 */
void kill_infinite_triangles (void)
{
  int e, t;

  /* First, mark all current triangles which are part of the convex hull. */
  
  trist_for (t)
    {
      e = EdFacet (t, 0);
      
      if (not search_is_infinite_triangle (e))
	if (search_is_infinite_triangle (Fnext (e)))
	  trist_hull_facet_set (e, TRUE);
	else if (search_is_infinite_triangle (Fnext (Sym (e))))
	  trist_hull_facet_set (Sym (e), TRUE);
    }

  /* Next, delete all infinite triangles. */
  
  trist_for (t)
    {
      e = EdFacet (t, 0);
      
      if (search_is_infinite_triangle (e)) {
	if (trist_hull_triangle (t))
	  trist_hull_facet_set (e, FALSE);
	trist_delete (EdFacet (t, 0));
      }
    }

}

/*--------------------------------------------------------------------------*/
/*
 * Routines for implementing no search.  In this case, all that really needs
 *          to be done is to delete the triangles removed by the flip.
 */

static void nsearch_init_func (int vertex[], int n)
{
  /* Do nothing! */
}

/*--------------------------------------------------------------------------*/

static int nsearch_search_func (int v)
{
  print ("Called nsearch_search_func -- this shouldn't happen!\n");
  Assert_always(FALSE);
}

/*--------------------------------------------------------------------------*/

static void nsearch_update_insert (int v, int ijk, const int field[], 
				   int i, int j, int k, int l)
{
  /* Do nothing! */
}

/*--------------------------------------------------------------------------*/

static void nsearch_update_triangle (int x, int y, int z, int o, int p,
				     const int field[/*5*/])
{
  /* Delete old triangles */

  trist_delete (field[0]);
}

/*--------------------------------------------------------------------------*/

static void nsearch_update_edge (int x, int y, int z, int o, int p,
				 const int field[/*6*/])
{
  /* Delete old triangles */

  trist_delete(field[0]);
  trist_delete(field[1]);
  trist_delete(field[2]);
}

/*--------------------------------------------------------------------------*/

static void nsearch_update_vertex (const int field[/*9*/])
{
  int i;

  /* Delete old triangles */

  upfor (i, 0, 5)
    trist_delete (field[i]);
}

/*--------------------------------------------------------------------------*/

static void nsearch_close (void)
{
  /* Do nothing! */
}



/*--------------------------------------------------------------------------*/
/**************** Functions for computing statistics ************************/

/*
 * infinite_num () -- This function will compute the number of infinite
 *		      facets.
 */
Infinite_num infinite_num (Trist_num tq)
{
  register int i, j, e;
  register int ext_f, ext_t;
  int ext_e;
  Infinite_num iq;
  ;
  ext_t = ext_f = 0;
  trist_for (j)
    if (is_current_triangle (j)) {
      if (outer_triangle (j))
	ext_f ++;
      upfor (i, EdFacet (j, 0), EdFacet (j, 1))
	if (    (not trist_hull_facet (i))
	    and (j == TrIndex (trist_tetra_min_ef (i)))
	    and (outer_tetra (i)))
	  ext_t ++;
    }
  ;

#ifndef __DEBUG__
  /* get edge numbers from Euler formula (ie equalities (1) and (3)) */
  e = (tq.v - 4) + (tq.f - ext_f) - (tq.t - ext_t) - 1;
  ext_e = tq.e - e;
#else
  /* use time-consuming count_edges() to compute them */
  count_inf_edges (&ext_e);
#endif
  iq.ext_e = ext_e;
  iq.ext_f = ext_f;
  iq.ext_t = ext_t;

  return (iq);
}

/*--------------------------------------------------------------------------*/

/*
 * search_info () -- This function returns a structure containing information
 *                   about the history dag structure and search time.
 */
Search_info search_info (int n)
{
  register int t, hier_size = 0;
  Search_info si;

  if (stype == TRIDAG_SEARCH) {
    si.avg_depth = (float) tot_vis / n;
    si.height = search_dag_height ();
  }

  if (not_inserted == 0)
    si.avg_depth_ninsert = 0;
  else
    si.avg_depth_ninsert = (float) vis_not_inserted / not_inserted;
  si.not_inserted = not_inserted;
  si.in_tests = in_tests;

  trist_for (t)
    if (not is_current_triangle (t))
      hier_size ++;

  si.hier_size = hier_size;
  return (si);
}


/*--------------------------------------------------------------------------*/

static void count_inf_edges (int *ext_e)
     /* infinite_num() refinement */
{
  register int i, j, ei, i1, i2;
  register int e, eh, et;
  char *aux;     /* used as bit vector to count edges! */
  int *dope;     /* we also need some dope man! */
  int max_index;
  ;
  dope = new_edge_dope (&max_index);
  aux = MALLOC (char, (max_index trist_Div8) + 1);
  BZERO (aux,   char, (max_index trist_Div8) + 1);
  ;
  et = eh = 6;

  e = 0;
  trist_for (j)
    if (is_current_triangle (j))
      upfor (i, EdFacet (j, 0), EdFacet (j, 5))  /* 50% useless !?!? */
	{
	  ei = edge_index (i, dope);
	  i1 = ei trist_Div8;
	  i2 = ei trist_Mod8;
	  if (not basic_charbit_on (i2, aux[i1]))
	    {
	      if (outer_edge(i))
		et ++;
	      else
		e ++;
	      aux[i1] = basic_charbit_s1 (i2, aux[i1]);
	    }
	}
  ;
  FREE (aux);
  kill_edge_dope (dope);
  *ext_e = et;
}

/*--------------------------------------------------------------------------*/

/**  Functions used in count_inf_edges()	**/

static int edge_index (int e, int dope[])
     /* returns an unique index for the edge given by edge facet e;
        dope must have been allocated via new_edge_dope();
        NOTE: there might be overlow if max_org too large! */
{
  int a = Org (e), b = Dest (e), aux;
  if (b < a)
    {
      aux = a;
      a = b;
      b = aux;
    }
  if ((b == a) or (b > trist_max_org ()))
    basic_error ("edge_index: Schu bi du.");
  return (dope[a] + b + 1);
}

/*--------------------------------------------------------------------------*/

static int *new_edge_dope (int *max_index)
{
  int i, m = trist_max_org ();
  int *dope;
  dope = MALLOC (int, m);
  dope[0] = -1;
  upfor (i, 1, m - 1)
    dope[i] = dope[i-1] + m - i;
  if (max_index)
    *max_index = dope[m-1] + m + 1;
  return (dope);
}

/*--------------------------------------------------------------------------*/

static void kill_edge_dope (int *dope)
{
  FREE (dope);
}

/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

/*
 * outer_tetra(ijk) -- returns TRUE or FALSE based on whether the tetrahedra
 *                        containing ijk is an outer tetra or an inner one.
 */
static int outer_tetra (int ijk)
{
  int i, j, k, l, i2, j2;		/* The vertices of the tetrahedra */

  trist_triangle (ijk, &i, &j, &k);
  trist_triangle (Fnext(ijk), &i2, &j2, &l);

  /* The outer vertices have index less than 5.  If any of the vertices
     are an outer vertex, then this is an outer tetrahedra                   */

  return (   is_infinite (i) or is_infinite (j)
	  or is_infinite (k) or is_infinite (l));
}


/*
 * outer_triangle(tri) -- returns TRUE or FALSE based on whether the
 *                           triangle ijk is an outer triangle or an inner one.
 */
static int outer_triangle (int tri)
{
  int x, y, z;

  /* The outer vertices have index less than 5.  If any of the vertices
     are an outer vertex, then this is an outer tetrahedra                   */

  trist_triangle (EdFacet(tri, 0), &x, &y, &z);
  return (is_infinite (x) || is_infinite (y) || is_infinite (z));
}

/*
 * outer_edge(e) -- returns TRUE or FALSE based on whether the edge e
 *                     contains a point at infinity.
 */
static int outer_edge (int e)
{
  return (is_infinite (Org(e)) or is_infinite (Dest(e)));
}

/*-------------------------------------------------------------------------*/

void print_search_dag(void)
{
  printf("\nSearchDag:\n\n");
  print_search_helper(1, 0);
}

/*-------------------------------------------------------------------------*/

static void print_search_helper(int t, int level)
{
  int i, x, y, z;

  if (t > 0) {
    for (i = 0; i <= 3 * level; i++)
      printf(" ");
    trist_triangle (EdFacet (t, 0), &x, &y, &z);
    printf("triangle %d = (%d, %d, %d)\n", t, x, y, z);
    for (i = 0; i <= 3 * level; i++)
      printf(" ");
    printf("Positive:\n");
    print_search_helper (pos_search (t), level + 1);
    for (i = 0; i <= 3 * level; i++)
      printf(" ");
    printf("Negative:\n");
    print_search_helper (neg_search (t), level + 1);
  }
}

/*------------------------------------------------------------------------*/

#define LEFT	    	-10
#define RIGHT       	-9
#define DONE        	-8
#define parent(x)   	(opposite_vertex (EdFacet(x, 0)))
#define set_parent(x,y)	(set_opposite_vertex(EdFacet(x, 0), y))
#define state(x)    	(pos_search (x))
#define set_state(x,y)	(set_pos_search(x, y))
#define height(x)   	(opposite_vertex (EdFacet(x, 1)))
#define set_height(x,y)	(set_opposite_vertex(EdFacet(x, 1), y))
#define not_visited(x) 	(state(x) > DONE)
#define max(x,y)	(((x) > (y)) ? (x) : (y))
  
/*------------------------------------------------------------------------*/

/*
 * search_dag_height() --
 *      Returns the height of the history dag.  Note that this procedure will
 *      destroy the history dag, so it should be used only after all other
 *      processing is done.  This function uses a DFS to compute the heights
 *      of each of the vertices.
 */
int search_dag_height (void)
{
  int t = 1, next_t, p_t;

  set_parent(t, -1);

  t = 1;

  while (t > 0) {
    
    switch state(t) {
    case RIGHT:
      next_t = neg_search (t);
      set_state(t, DONE);
      if (next_t > 0)
	if (not_visited(next_t)) {
	  set_parent(next_t, t);
	  t = next_t;
	}
	else {
	  set_height (t, max (height(t), height(next_t) + 1));
	}

      break;
    case DONE:
      p_t = parent(t);
      if (p_t > 0) {
	set_height (parent(t), max (height(parent(t)), height(t) + 1));
      }

      t = parent(t);
      break;
    default:
      set_height(t, 0);
      next_t = pos_search (t);   /* NOTE: these two statements need to be in */
      set_state(t, RIGHT);       /*       this order.                        */
      if (next_t > 0)
	if (not_visited(next_t)) {
	  set_parent(next_t, t);
	  t = next_t;
	}
	else {
	  set_height (t, max (height(t), height(next_t) + 1));
	}
    }
  }

  

  return (height(1));
}
     
/*-----------------------------------------------------------------------*/

#ifdef ONLY_FOR_TESTING

static void single_test(int i, int j, int k, int l)
{
  print("sos_positive3_inf(%d, %d, %d, %d) = %d\n", i, j, k, l, 
	sos_positive3_inf(i,j,k,l));
}

static void test_infinite(void)
{
  int i, j, inf = outer_index;

  print("\n\nTests for positive3_inf1\n");

  single_test(inf, 5, 6, 7);
  single_test(inf+1, 5, 6, 7);
  single_test(inf+2, 5, 6, 7);
  single_test(inf+3, 5, 6, 7);
  single_test(inf, 6, 5, 7);
  single_test(inf+1, 6, 5, 7);
  single_test(inf+2, 6, 5, 7);
  single_test(inf+3, 6, 5, 7);

  single_test(inf, 5, 6, 8);
  single_test(inf+1, 5, 6, 8);
  single_test(inf+2, 5, 6, 8);
  single_test(inf+3, 5, 6, 8);
 
  print("\nTests for positive3_inf2\n");

  single_test(inf+0, inf+1, 5, 6);
  single_test(inf+0, inf+2, 5, 6);
  single_test(inf+0, inf+3, 5, 6);
  single_test(inf+1, inf+2, 5, 6);
  single_test(inf+1, inf+3, 5, 6);
  single_test(inf+2, inf+3, 5, 6);

  print("\nTesting for coplanar points\n");

  for (i = 11; i <= 14; i++) {
    print("\n\tFinite point %d\n", i);
    for (j = inf; j <= inf+4; j++)
      single_test (j, 9, 10, i);
  }

  print ("\n\nTesting three infinite points\n");

  single_test (5, inf+0, inf+1, inf+2);
  single_test (5, inf+0, inf+1, inf+3);
  single_test (5, inf+0, inf+2, inf+3);
  single_test (5, inf+1, inf+2, inf+3);
  single_test (9, inf+0, inf+1, inf+2);
  single_test (9, inf+0, inf+1, inf+3);
  single_test (9, inf+0, inf+2, inf+3);
  single_test (9, inf+1, inf+2, inf+3);

  print ("\nTests for 4 infinite points\n\n");

  single_test (inf, inf+1, inf+2, inf+3);
  single_test (inf+1, inf+3, inf+2, inf);
}

#endif
