/* delcx/stack.c --- Unbounded int stack with dynamic re-allocation. 
                      Modified from basic/istaque.c written by Ernst Mucke.
                      All calls to the basic/malloc routines are removed.
                      They slow it down too much if there are many stacks 
                      in existence.                                         */

/*--------------------------------------------------------------------------*/

#include "delcx.h"

/*--------------------------------------------------------------------------*/

/* Local macros for calling malloc and realloc.  These avoid calling the
   basic routines so that long searches don't occur in malist_delete.       */

#define MMALLOC(T,N) \
  (T*) malloc ((int) sizeof(T) * (N))

#define RREALLOC(P,T,N) \
  P = (T*) realloc ((char*) P, (int) sizeof(T) * (N))

/*--------------------------------------------------------------------------*/

#define DO_TYPECHECKING  /* Uncomment this for no run-time type checking. */
#define PRINT_INTERNALS  /* See casc_istaque_print(). */

#define Magic  130862002

#define default_factor  ((double) 2.0)
#define default_shrink  FALSE
 
/* NOTE: It shrinks ONLY WHEN casc_istaque_pop is used.  The used shrinking
         scheme didn't work well for queue (get/push) operations.  In fact,
         automatic shrinking might be always bad, therefore the default. */

#if defined (__DEBUG__) && (! defined (DO_TYPECHECKING))
# define DO_TYPECHECKING
#endif

/*--------------------------------------------------------------------------*/

typedef struct Int_stack_type
{
  int magic;
  int *stack, top, stacksize;
  int *bottom, head, tail, botsize;
  double factor;
  int minsize;
  Basic_byte shrink;
} Int_Stack;

#ifdef DO_TYPECHECKING
 static Int_Stack *minp = (Int_Stack *) 0xffffffff;
 static Int_Stack *maxp = (Int_Stack *) 0x00000000;
#endif

static top(), bot();

#define non_empty_stack(Q)  ((Q)->top >= 0)
#define non_empty_bottom(Q)  ((Q)->head < (Q)->tail)
#define clear(Q) do { Q->top = Q->head = Q->tail = -1; } while(0)

/*---------------------------------------------------------------------------*/

/* Statistics for stack usage */

int current_storage = 0;
int max_storage = 0;

/*---------------------------------------------------------------------------*/

/* Local Functions */

static Int_Stack* cast (Casc_Stack_adt s);
static int top (Int_Stack *q);
static int bot (Int_Stack *q);

/*---------------------------------------------------------------------------*/

#ifdef DO_TYPECHECKING

static Int_Stack* cast (Casc_Stack_adt s)
     /* This casts the "abstract data type" Casc_Stack_adt, ie (char *),
        to (Int_Stack *).  Type checking is done at runtime by
        checking a magic number and minp & maxp. */
{
  Int_Stack *q = (Int_Stack *) s;
  Assert_always (q and (minp <= q) and (q <= maxp));
  Assert_always ((q->magic == Magic));
  return (q);
}

#else

#define cast(S)  ((Int_Stack *) (S))

#endif

/*--------------------------------------------------------------------------*/

Casc_Stack_adt casc_stack_new (int minsize)
     /* Allocates memory for a new object of Casc_Stack_adt.
        Use casc_istque_dispose() to free the memory. */
{
  Int_Stack *q;
  if (minsize < 1)
    basic_error ("casc_stack_new: invalid minsize");
  q = MMALLOC (Int_Stack, 1);
  q->magic = Magic;
  q->stack = MMALLOC (int, minsize);
  q->bottom = MMALLOC (int, 1);  /* just to initialize; why not = NULL ? */
  q->stacksize = q->minsize = minsize;
  q->botsize = 0;
  clear (q);
  q->factor = default_factor;
  q->shrink = default_shrink;
#ifdef DO_TYPECHECKING
  minp = (Int_Stack *) Min (minp, q);
  maxp = (Int_Stack *) Max (maxp, q);
#endif

  /* Update storage info */

  current_storage += sizeof(Int_Stack) + (minsize + 1) * sizeof (int);
  if (current_storage > max_storage)
    max_storage = current_storage;

  return ((Casc_Stack_adt ) q);
}

/*--------------------------------------------------------------------------*/

void casc_stack_dispose (Casc_Stack_adt s)
     /* De-allocates the memory of s.
        NOTE that this does not zero the value of s; ergo what you 
        really want to do is: "casc_stack_dispose (s); s = NULL". */
{
  Int_Stack *q = cast (s);

  /* Update storage info */

  current_storage -= sizeof(Int_Stack) + (q->stacksize + 1) * sizeof (int);

  FREE (q->stack);
  FREE (q->bottom);
  FREE (q);
}

/*--------------------------------------------------------------------------*/

void casc_stack_clear (Casc_Stack_adt s)
     /* Clears the contents of s. */
{
  Int_Stack *q = cast (s);
  clear (q);
}

/*--------------------------------------------------------------------------*/

int casc_stack_empty (Casc_Stack_adt s)
     /* Tests if s is empty. */
{
  Int_Stack *q = cast (s);
  return (not (non_empty_stack (q) or non_empty_bottom (q)));
}

/*--------------------------------------------------------------------------*/

int casc_stack_top (Casc_Stack_adt s)
     /* Returns the top element (stack operation!) but does not pop it. */
{
  Int_Stack *q = cast (s);
  return (top (q));
}

static int top (Int_Stack *q)
{
  if (non_empty_stack (q))
    return (q->stack[q->top]);
  else if (non_empty_bottom (q))
    return (q->bottom[q->tail-1]);
  else
    return (0 - MAXINT);
}

/*--------------------------------------------------------------------------*/

int casc_stack_pop (Casc_Stack_adt s)
     /* Returns the top element (stack operation!) and pops it. */
{
  Int_Stack *q = cast (s);
  int result = top (q);
  if (non_empty_stack (q))
    { 
      q->top --;
      if ((q->shrink) and (q->factor > 1) and (q->stacksize > q->minsize)
          and (q->top < q->stacksize / q->factor / q->factor))
        { /* shrink stack */
          int newsize = (int) (q->stacksize / q->factor);
          if (newsize < q->minsize) newsize = q->minsize;
          RREALLOC (q->stack, int, newsize);

	  /* Update storage info */

	  current_storage -= (q->stacksize - newsize) * sizeof (int);

          q->stacksize = newsize;
        }
    }
  else if (non_empty_bottom (q))
    q->tail --;    /* we never shrink the bottom */
  else
    basic_error ("casc_stack_pop: underflow");
  return (result);
}

/*--------------------------------------------------------------------------*/

void casc_stack_push (Casc_Stack_adt s, int value)
     /* Adds a new top element (stack and queue operation!). */
{
  Int_Stack *q = cast (s);
  q->top ++;
  if (q->top == q->stacksize)
    if (q->factor <= 1)
      basic_error ("casc_stack_push: overflow");
    else
      { /* make stack larger */
        int newsize = (int) (0.5 + q->factor * q->stacksize);
        RREALLOC (q->stack, int, newsize);

	/* Update storage info */

	current_storage += (newsize - q->stacksize) * sizeof (int);
	if (current_storage > max_storage)
	  max_storage = current_storage;

        q->stacksize = newsize;
      }
  q->stack[q->top] = value;
}

/*--------------------------------------------------------------------------*/

int casc_stack_bot (Casc_Stack_adt s)
     /* Returns the bottom element (queue operation!)
        but does not remove it. */
{
  Int_Stack *q = cast (s);
  return (bot (q));
}

static int bot (Int_Stack *q)
{
  if (non_empty_bottom (q))
    return (q->bottom[q->head]);
  else if (non_empty_stack (q))
    return (q->stack[0]);
  else
    return (0 - MAXINT);
}

/*--------------------------------------------------------------------------*/

int casc_stack_get (Casc_Stack_adt s)
     /* Returns the bottom element (queue operation!) and removes it. */
{
  Int_Stack *q = cast (s);
  int result = bot (q);
  if (non_empty_bottom (q))
    q->head ++;
  else if (non_empty_stack (q))
    { /* move stack to bottom */
      int *aux = q->bottom, auxsize = q->botsize;
      q->bottom = q->stack;
      q->head = 1;
      q->tail = q->top + 1;
      q->botsize = q->stacksize;
      q->stack = aux;
      if (q->stacksize <= auxsize)
        q->stacksize = auxsize;
      else
        {
          RREALLOC (q->stack, int, q->stacksize);
        }
      q->top = -1;
    }
  else
    basic_error ("casc_stack_get: underflow");
  return (result);
}

/*--------------------------------------------------------------------------*/

int casc_stack_length (Casc_Stack_adt s)
     /* Returns the number of elements in s. */
{
  Int_Stack *q = cast (s);
  int len = 0;
  if (non_empty_stack (q))
    len += q->top + 1;
  if (non_empty_bottom (q))
    len += q->tail - q->head;
  return (len);
}

/*--------------------------------------------------------------------------*/

void casc_stack_print (Casc_Stack_adt s, FILE *file)
     /* Prints elements of s to given file. */
{
  Int_Stack *q = cast (s);
  int i;
#ifdef PRINT_INTERNALS
  fprint (file, "Casc_Stack_adt 0x%x-> (%d;0x%x,%d,%d;0x%x,%d,%d,%d) ",
          q, q->minsize,
          q->stack, q->top, q->stacksize,
          q->bottom, q->head, q->tail, q->botsize);
#endif
  if (casc_stack_empty (s))
    fprint (file, "empty.");
  if (non_empty_bottom (q))
    {
      fprint (file, "%d", q->bottom[q->head]);
      upfor (i, q->head + 1, q->tail - 1) fprint (file, ",%d", q->bottom[i]);
      if (non_empty_stack (q)) fprint (file, ";");
    }
  if (non_empty_stack (q))
    {
      fprint (file, "%d", q->stack[0]);
      upfor (i, 1, q->top) fprint (file, ",%d", q->stack[i]);
    }
}

/*--------------------------------------------------------------------------*/

double casc_stack_resize (Casc_Stack_adt s, double resize_code)
     /* Sets the new resize_code and returns the value of the old one.
        The resize factor is abs (resize_code).  If the resize_code is
        negative, shrinking is allowed as well.  The default resize_code
        is +2; meaning that the size of the list is doubled whenever it
        gets filled up.  In this case, the overhead of this semi-dynamic
        solution is O(1) in average. */ 
{
  Int_Stack *q = cast (s);
  double old = If (q->shrink, -q->factor, q->factor);
  q->factor = resize_code;
  q->shrink = FALSE;
  if (resize_code < 0)
    {
      q->factor = -q->factor;
      q->shrink = TRUE;
    }
  return (old);
}


/*--------------------------------------------------------------------------*/

int casc_max_storage (void)
{
  return (max_storage);
}

