/*
 * tridag.c -- Functions for implementing the history dag.
 *
 * Author -- Mike Facello
 *
 * Date -- 5/31/94 (many changes from old tridag.c)
 */

#include "delcx.h"
#include "dt.h"

/*--------------------------------------------------------------------------*/

#define mod2       & 01

/*--------------------------------------------------------------------------*/

/* Accessor functions for tridag/search records */
/* Here, t is a triangle index, e is an edfacet number */

int pos_search(int t)
{
  Dag_record *dg = (Dag_record *) trist_data_addr (t);
  
  return (dg->search[0]);
}

/*-------------------------*/

int neg_search(int t)
{
  Dag_record *dg = (Dag_record *) trist_data_addr (t);

  return (dg->search[1]);
}

/*-------------------------*/

void set_pos_search(int t1, int t2)
{
  Dag_record *dg = (Dag_record *) trist_data_addr (t1);

  dg->search[0] = t2;
}

/*-------------------------*/

void set_neg_search(int t1, int t2)
{
  Dag_record *dg = (Dag_record *) trist_data_addr (t1);

  dg->search[1] = t2;
}

/*-------------------------*/

int search_ptr(int e)
{
  register int t = TrIndex(e);
  Dag_record *dg = (Dag_record *) trist_data_addr (t);

  return (dg->search[e mod2]);
}

/*-------------------------*/

void set_search_ptr(int e, int t)
{
  register int t1 = TrIndex (e);
  Dag_record *dg = (Dag_record *) trist_data_addr (t1);
  
  dg->search[e mod2] = t;
}

/*-------------------------*/
/* Routines used by the regular triangulation program only */

#ifdef REGULAR

Trist_org opposite_vertex(int e)
{
  register int t = TrIndex(e);
  Dag_record *dg = (Dag_record *) trist_data_addr (t);

  return (dg->opposite[e mod2]);
}

/*-------------------------*/

void set_opposite_vertex(int e, Trist_org v)
{
  register int t = TrIndex (e);
  Dag_record *dg = (Dag_record *) trist_data_addr (t);
  
#ifdef SDEBUG
  if (v > 0) {
    int i,j,k;
    trist_triangle (e, &i, &j, &k);
    Assert_always (sos_positive3_inf (v, i, j, k));
  }
#endif

  dg->opposite[e mod2] = v;
}

#endif

/*-------------------------*/

#ifdef SDEBUG

void set_flip_type (int t, int type)
{
  Dag_record *dg = (Dag_record *) trist_data_addr (t);

  if (type < 0)
    dg->flip_type = -dg->flip_type;
  else
    dg->flip_type = type;
}

/*-------------------------*/

void print_flip_type (int t)
{
  int type;
  Dag_record *dg = (Dag_record *) trist_data_addr (t);

  if (dg->flip_type < 0) {
    print ("vertex and ");
    type = -dg->flip_type;
  }
  else
    type = dg->flip_type;

  switch (type) {
  case F_TRIANGLE:
    print ("triangle(1)");
    break;
  case F_EDGE:
    print ("edge(0)");
    break;
  case F_VERTEX:
    print ("vertex");
    break;
  case F_INSERT:
    print ("insert(2)");
  }
}

#endif /* SDEBUG */

/*--------------------------------------------------------------------------*/

/*
 * int is_current_triangle (t) -- Returns true it t is in the search
 *          hierarchy but IS NOT in the current triangulation.  Note that
 *          there are some triangles in the hierarchy that are also in
 *          the current triangulation.  This predicate is false for those.
 */
int is_current_triangle (int t)
{
  if (trist_deleted(t))
    return (FALSE);
  else
    return (Fnext(EdFacet(t, 0)) != EdFacet(t, 0));
}

/*--------------------------------------------------------------------------*/

/*
 * tridag_make (i, j, k) -- creates a new triangle and initializes history
 *			    dag fields
 */

int tridag_make (int i, int j, int k)
{
  int ijk = trist_make (i, j, k);
  
  set_search_ptr (ijk, -1);
  set_search_ptr (Sym (ijk), -1);
  set_opposite_vertex (ijk, 0);
  set_opposite_vertex (Sym (ijk), 0);

  return (ijk);
}

/*--------------------------------------------------------------------------*/

/*
 * tridag_kill -- This function deletes all search triangles that are not
 *                part of the final triangulation.
 */
void tridag_kill (void)
{
  int t;

  trist_for (t)
    if (not is_current_triangle (t))
      trist_delete (EdFacet (t, 0));
}

/*-----------------------------------------------------------------------*/

void tridag_init (int vertex[], int n)
{
  /* Nothing to really do here, just a dummy function for search.c */
}

/*-----------------------------------------------------------------------*/

/*
 * tridag_locate_vertex (v) --  Finds the tetrahedron that v lies in, 
 *                              and returns a triangle of that tetrahedron.
 */
int tridag_locate_vertex (int v)
{
  int t, next_t, orient;
  int done;
  int i,j,k,l;
  int vis = 0; 		 /* Counter for # of nodes visited */

  t = 1;       /* Search always starts at triangle 1 */
  done = FALSE;

  while (!done) {
    int cur_t = EdFacet(t, 0);

    /* pos orient = 1, neg orient = 0 */
    orient = search_pos_orientation (v, cur_t);  
                                     /* ^^^^^ used to be just t  */

#ifdef SDEBUG
    trist_triangle (cur_t, &i, &j, &k);
    printf ("Orientation with triangle (%d, %d, %d) is %s", i, j, k,
	    (orient ? "positive" : "negative"));
    print (" with flip type:  ");
    print_flip_type (t);
    print ("\n");
#endif


#ifdef REGULAR

    /* The following code will make sure that the point being inserted is
       inside the sphere of the current tetrahedron (i.e. it is still
       on the lifted convex hull).  This is only done if we are currently
       inserting this vertex (indicated by boolean do_insert).             */

    {
      int test;
      int o;
      int tri = EdFacet(t, (orient ? 0 : 1));
      int i,j,k;

      trist_triangle(tri, &i, &j, &k);
      if (    ((vis % redcheck_dist) == 0)
	  and ((o = opposite_vertex(tri)) > 0)) {
	in_tests ++;
#ifdef SDEBUG
	print ("Testing that point is inside tetrahedron (%d, %d, %d, %d)\n",
	       i, j, k, o);
	Assert_always (    sos_positive3_inf (v,i,j,k)
		       and sos_positive3_inf (v,j,i,o)
		       and sos_positive3_inf (v,k,j,o)
		       and sos_positive3_inf (v,i,k,o));
#endif
	if (not sos_in_sphere_inf_p(i, j, k, o, v)) {
#ifdef SDEBUG
	  print ("&&&&& Not inserting vertex %d\n", v);
#endif
	  tot_vis += vis;
	  vis_not_inserted += vis;
	  not_inserted++;   /* This point is "flipped out" of the triang.  */
	  add_redundant_vertex(v);
	  return (-1);
	}
      }
    }
#endif

    if (orient)
      next_t = pos_search(t);
    else
      next_t = neg_search(t);

    if (next_t > -1)
      t = next_t;
    else
      done = TRUE;

    vis++;
  }

  /* If the vertex is not inside the sphere formed by the others, it
     should not be inserted.                                         */

#ifdef REGULAR
  trist_tetra(EdFacet(t, (orient ? 0 : 1)), &i, &j, &k, &l);
  in_tests ++;
  if (not (sos_in_sphere_inf_p (i,j,k,l,v))) {
#ifdef SDEBUG
    print ("&&&&& Not inserting vertex %d\n", v);
#endif
    not_inserted ++;
    vis_not_inserted += vis;
    add_redundant_vertex(v);
    return (-1);
  }
#endif
  tot_vis += vis;


  return (EdFacet(t, (orient ? 0 : 1)));
}

/*-----------------------------------------------------------------------*/

/*
 * tridag_locate_redundant (v) --  Finds the tetrahedron that redundant
 *                                 vertex v lies in, and returns a triangle
 *                                 of that tetrahedron.
 */
int tridag_locate_redundant (int v)
{
  int t, next_t, orient;
  int done;
  int i,j,k,l;

  t = 1;       /* Search always starts at triangle 1 */
  done = FALSE;

  while (!done) {
    int cur_t = EdFacet(t, 0);

    /* pos orient = 1, neg orient = 0 */
    orient = search_pos_orientation (v, cur_t);  
                                     /* ^^^^^ used to be just t  */

#ifdef SDEBUG
    trist_triangle (cur_t, &i, &j, &k);
    printf ("Orientation with triangle (%d, %d, %d) is %s", i, j, k,
	    (orient ? "positive" : "negative"));
    print (" with flip type:  ");
    print_flip_type (t);
    print ("\n");
#endif

    if (orient)
      next_t = pos_search(t);
    else
      next_t = neg_search(t);

    if (next_t > -1)
      t = next_t;
    else
      done = TRUE;
  }

  return (EdFacet(t, (orient ? 0 : 1)));
}

/*-----------------------------------------------------------------------*/

/*
 * tridag_update_insert () -- This function updates the history dag
 *                            for a vertex insertion.
 */
void tridag_update_insert (int v, int ijk, const int field[/*9*/], 
			   int i, int j, int k, int l)
{
  int ijv = field[0];
  int ikv = field[1];
  int jkv = field[2];
  int ilv = field[3];
  int jlv = field[4];
  int klv = field[5];
  int jil = field[6];
  int kjl = field[7];
  int ikl = field[8];

  int t = TrIndex(jlv);

  set_search_ptr (jlv, TrIndex (ilv));
  set_search_ptr (Sym(jlv), TrIndex (klv));
  set_search_ptr (ilv, TrIndex (ikv));
  set_search_ptr (Sym(ilv), TrIndex (ijv));
  set_search_ptr (klv, TrIndex (jkv));
  set_search_ptr (Sym(klv), TrIndex (ikv));
  set_search_ptr (ikv, -1);
  set_search_ptr (Sym(ikv), -1);
  set_search_ptr (ijv, -1);
  set_search_ptr (Sym(ijv), -1);
  set_search_ptr (jkv, -1);
  set_search_ptr (Sym(jkv), -1);

  if (search_ptr (ijk) == -1)
    set_search_ptr (ijk, t);
  if (search_ptr (jil) == -1)
    set_search_ptr (jil, t);
  if (search_ptr (kjl) == -1)
    set_search_ptr (kjl, t);
  if (search_ptr (ikl) == -1)
    set_search_ptr (ikl, t);

  set_opposite_vertex (ikv, j);
  set_opposite_vertex (Sym(ikv), l);
  set_opposite_vertex (ijv, l);
  set_opposite_vertex (Sym(ijv), k);
  set_opposite_vertex (jkv, l);
  set_opposite_vertex (Sym(jkv), i);

#ifdef SDEBUG
  print ("Setting FLIP types in insert\n");
  set_flip_type (TrIndex(jlv), F_INSERT);
  set_flip_type (TrIndex(ilv), F_INSERT);
  set_flip_type (TrIndex(klv), F_INSERT);
  set_flip_type (TrIndex(ikv), F_INSERT);
  set_flip_type (TrIndex(ijv), F_INSERT);
  set_flip_type (TrIndex(jkv), F_INSERT);
#endif

}


/*------------------------------------------------------------------------*/

/* tridag_update_triangle -- This procedure updates the search structure
 *      for a triangle flip.  The only search pointers that could be pointing
 *      to one of the deleted tetrahedra are ones from the surrounding
 *      faces.  The pointers of these faces are updated only if they point
 *      to this tetrahedra (i.e. pointer == -1).
 */
#ifdef OLDSTUFF
void tridag_update_triangle (int xyz, const int field[/*5*/], 
			     int x, int y, int z, int o, int p)
#endif
void tridag_update_triangle (int x, int y, int z, int o, int p,
			     const int field[/*10*/])
{
  /* Update the search structure */

  int j;
#ifdef OLDSTUFF
  int pox = Enext (Turn(field[2]));  /* field[2] = zxo */
#endif
  int pox, poy, poz;
  int t;

#ifdef SDEBUG
  printf("Flipping triangle %d = (%d, %d, %d)\n", field[0], x, y, z);
#endif

  t = TrIndex (pox);
  pox = Sym(Enext2(field[1]));
  poy = Sym(Enext2(field[2]));
  poz = Sym(Enext2(field[3]));
  Assert (dt_test_triangle(pox, p, o, x));
  Assert (dt_test_triangle(poy, p, o, y));
  Assert (dt_test_triangle(poz, p, o, z));
  set_search_ptr(pox, TrIndex (poy));
  set_search_ptr(Sym(pox), TrIndex (poz));
  set_search_ptr(poy, -1);
  set_search_ptr(Sym(poy), -1);
  set_search_ptr(poz, -1);
  set_search_ptr(Sym(poz), -1);

  set_opposite_vertex (poy, z);
  set_opposite_vertex (Sym(poy), x);
  set_opposite_vertex (poz, x);
  set_opposite_vertex (Sym(poz), y);

#ifdef SDEBUG
  set_flip_type(TrIndex(pox), F_TRIANGLE);
  set_flip_type(TrIndex(poy), F_TRIANGLE);
  set_flip_type(TrIndex(poz), F_TRIANGLE);
#endif

  /* update the triangle that was removed. */
  if (search_ptr (field[0]) == -1)
    set_search_ptr (field[0], t);
  if (search_ptr (Sym(field[0])) == -1)
    set_search_ptr (Sym(field[0]), t);

  /* The field array contains the 6 outward pointing triangles of the double
     tetrahedron for this flip.                                             */

#ifdef OLDSTUFF
  upfor (j, 0, 5)
    if (search_ptr (Sym(field[j])) == -1)
      set_search_ptr (Sym(field[j]), t);
#endif
  upfor (j, 4, 9)
    if (search_ptr (field[j]) == -1)
      set_search_ptr (field[j], t);
}


/*------------------------------------------------------------------------*/

/* tridag_update_edge -- This procedure updates the search structure
 *      for an edge flip.  The only search pointers that could be pointing
 *      faces.  The pointers of these faces are updated only if they point
 *      to this tetrahedra (i.e. pointer == -1).
 */
#ifdef OLDSTUFF
void tridag_update_edge (int pox, int poy, int poz, const int field[/*5*/],
			 int x, int y, int z, int o, int p)
#endif
void tridag_update_edge (int x, int y, int z, int o, int p,
			 const int field[/*5*/])
{
  /* Update the search structure */

  int j;
#ifdef OLDSTUFF
  int yxz = Fnext (Sym (field[0]));   /* field[0] = xyo */
#endif
  int yxz = Sym (field[3]);   /* field[3] = xyz */
  int t;
  int pox = field[0];
  int poy = field[1];
  int poz = field[2];


#ifdef SDEBUG
  printf("Flipping edge (%d, %d) with adjacent triangle %d, %d, and %d\n",
	 p, o, pox, poy, poz);
#endif

  t = TrIndex (yxz);
  Assert (dt_test_triangle(yxz, y, x, z));
  set_search_ptr (yxz, -1);
  set_search_ptr (Sym (yxz), -1);

  set_opposite_vertex (yxz, p);
  set_opposite_vertex (Sym(yxz), o);

#ifdef SDEBUG
  set_flip_type (TrIndex(yxz), F_EDGE);
#endif

  /* update the triangles that were removed. */
  if (search_ptr (pox) == -1)
    set_search_ptr (pox, t);
  if (search_ptr (Sym(pox)) == -1)
    set_search_ptr (Sym(pox), t);
  if (search_ptr (poy) == -1)
    set_search_ptr (poy, t);
  if (search_ptr (Sym(poy)) == -1)
    set_search_ptr (Sym(poy), t);
  if (search_ptr (poz) == -1)
    set_search_ptr (poz, t);
  if (search_ptr (Sym(poz)) == -1)
    set_search_ptr (Sym(poz), t);

#ifdef OLDSTUFF
  upfor (j, 0, 5)
    if (search_ptr (Sym(field[j])) == -1)
      set_search_ptr (Sym(field[j]), t);
#endif

  upfor (j, 4, 9)
    if (search_ptr (field[j]) == -1)
      set_search_ptr (field[j], t);
}


/*--------------------------------------------------------------------------*/

/*
 * tridag_update_vertex (field) --
 *  	   After a vertex flip has occured, this procedure will update the
 *         search pointers in the point location structure.
 */
void tridag_update_vertex (const int field[/*9*/])
{
  int i, leaf = -1;

#ifdef DEBUG
  print ("%d ", Dest(field[0]));
#endif

  upfor (i, 6, 9) {

    /* Reset all opposite pointers, because we don't know what they are now */
    set_opposite_vertex(field[i], 0);

    if (search_ptr (field[i]) == -1) {
      leaf = TrIndex (field[i]);

#ifdef SDEBUG
      set_flip_type(TrIndex(field[i]),-1);
#endif
    }
  }

  if (leaf == -1) {
    int x, y, z, new_tri;
    int s, t;

#ifdef DEBUG
    print ("Creating new triangle\n");

    pr_triangle(field[6]);
    pr_triangle(Fnext(field[6]));
    pr_triangle(Fnext(Sym(field[6])));
#endif

    trist_triangle (field[6], &x, &y, &z);
    trist_triangle (Fnext(field[6]), &x, &y, &s);
    trist_triangle (Fnext(Sym(field[6])), &y, &x, &t);
    
    new_tri = tridag_make (x, y, z);

#ifdef SDEBUG
    set_flip_type (TrIndex(new_tri), F_VERTEX);
#endif


    /* Insert the new triangle into the structure in place of field[6] */
    
    Fsplice(new_tri, field[6]);
    Fsplice(Enext(new_tri), Enext(field[6]));
    Fsplice(Enext2(new_tri), Enext2(field[6]));
    
    delete_the_triangle(field[6]);
    
    (void) dt_test_open_tetra (new_tri, x, y, t, z);
    (void) dt_test_open_tetra (Sym(new_tri), y, x, s, z);

    leaf = TrIndex (new_tri);
  }

  upfor (i, 0, 5) {
    if (search_ptr (field[i]) == -1)
      set_search_ptr (field[i], leaf);
    if (search_ptr (Sym(field[i])) == -1)
      set_search_ptr (Sym(field[i]), leaf);
  }
}




