/* detri/search.c --- Simple but efficient point location in (Delaunay) Tri. */

/*--------------------------------------------------------------------------*/

#include "trist.h"

#ifdef __DEBUG__
 int dt_test_open_tetra (int a, int x, int y, int z, int o);
 int dt_test_triangle (int ef, int x, int y, int z);
#endif

/*--------------------------------------------------------------------------*/

/* locals */

static double distance (int ef, int p);
static double square (double x);

/* to keep track of work done */
static int positive3_tests = 0;
static int tets_visited = 0;

/*--------------------------------------------------------------------------*/

int trist_search (int p, int k)
     /* Point location for query point p in (Delaunay) triangulation of
        current Trist.  Uses jump'n'walk strategy, with a random sample of
        size k.  Use k=O(n^{1/4}), n == num. of points in triangulation; to
        get an expected search time of O(n^{1/4}).
         Assumes: k > 0 and srandom (seed).
         The result is an edgefacet ef with p in ef^+.  If it's a hull facet,
        then p lies outside the convex hull; otherwise, the point lies inside
        the tetrahedron represented by ef. */
{
  int result, a, b, c, i, ef, ef_prime, last = trist_current ()->last_triangle;
  double dist, dist_prime;
  ;
#define GET_RANDOM_FACET(X)  /* Input/output: int x. */               \
  do                                                                   \
    {                                                                   \
      X = ((int) random () mod last) + 1; /* Result: random triangle */  \
    } until (not trist_deleted (X));                                      \
  X = EdFacet (X, 0)
  ;
  /* Jump: Select "good" candidate using k random samples, taking the closest
   *       one useing quick'n'dirty floating-point distance()
   */
  GET_RANDOM_FACET (ef_prime);
  dist_prime = distance (ef_prime, p);
  upfor (i, 2, k)
    {
      GET_RANDOM_FACET (ef);
      dist = distance (ef, p);
      if (dist < dist_prime)
        {
          ef_prime = ef;
          dist_prime = dist;
        }      
    }
  trist_triangle (ef_prime, &a, &b, &c);
  if (not sos_positive3 (p, a, b, c))
    {
      ef_prime = Sym (ef_prime);
      trist_triangle (ef_prime, &a, &b, &c);
    }
  positive3_tests ++;
  ;
  /* Walk: start search at "good" candidate, "walk" towards target,
   *       and return result.
   */
  result = trist_search_walk (p, ef_prime, a, b, c);
  ;
#if __DEBUG__
  trist_triangle (result, &a, &b, &c);
  if (trist_hull_facet (result))
    /* p is on CH and visible from v, right? */
    Assert (sos_positive3 (p, a, b, c));
  else
    { /* p is inside tetrahedron ef^+, right? */
      i = Dest (Enext (Fnext (result)));
      (void) dt_test_open_tetra (Fnext (result), a, b, c, i);
      Assert (    sos_positive3 (p, a, b, c)
              and sos_positive3 (p, b, a, i)
              and sos_positive3 (p, c, b, i)
              and sos_positive3 (p, a, c, i));
      }
#endif
  return (result);
}

/*--------------------------------------------------------------------------*/

int trist_search_walk (int p, int ef, int i, int j, int k)
     /* The "walk" routine. -- Right now implemented with tail recursion! --
        It locates point p within (Delaunay) triangulation of current Trist,
        assuming ef == (i,j,k) and sos_positive3 (p, i, j, k) == TRUE. 
        The result... see trist_search(). */
     /*
      * NOTE: This still needs some modification to make it work for
      *       non-Delaunay triangulations with "high" probability!
      */
{
  Assert (dt_test_triangle (ef, i, j, k) and sos_positive3 (p, i, j, k));
  if (trist_hull_facet (ef))
    {
      return (ef);   /* Found!  Outside convex hull. */
    }
  else
    {
      int ef_1 = Fnext (ef);
      int ef_2 = Fnext (ef = Enext (ef));
      int ef_3 = Fnext (ef = Enext (ef));
      int o = Dest (Enext (ef_1));
      tets_visited ++;
      if (sos_positive3 (p, i, j, o))
        {
          positive3_tests += 1;
          return (trist_search_walk (p, ef_1, i, j, o));
        }
      else if (sos_positive3 (p, j, k, o))
        {
          positive3_tests += 2;
          return (trist_search_walk (p, ef_2, j, k, o));
        }
      else if (sos_positive3 (p, k, i, o))
        {
          positive3_tests += 3;
          return (trist_search_walk (p, ef_3, k, i, o));
        }
      else
        {
          /* Found! Inside tetrahedron. */
          positive3_tests += 3;
          return (ef);    
        }
    }
}

/*--------------------------------------------------------------------------*/

static double distance (int ef, int p)
     /* Returns FP "distance" of edfacet ef to p. */
{
  int a, b, c;
  double d, da, db, dc;
  double p1 = sos_fp (p, 1);
  double p2 = sos_fp (p, 2);
  double p3 = sos_fp (p, 3);
  trist_triangle (ef, &a, &b, &c);
  da = (  square (p1 - sos_fp (a, 1))
        + square (p2 - sos_fp (a, 2))
        + square (p3 - sos_fp (a, 3)));
  db =  (  square (p1 - sos_fp (b, 1))
         + square (p2 - sos_fp (b, 2))
         + square (p3 - sos_fp (b, 3)));
  dc = (  square (p1 - sos_fp (c, 1))
        + square (p2 - sos_fp (c, 2))
        + square (p3 - sos_fp (c, 3)));
  d = Min (da, db);
  return (Min (d, dc));
}

#if 0

/* NOTE: Below "distance" function is slightly cheaper,
         but yields to worse results. */

static double distance (int ef, int p)
     /* Returns FP "distance" of edfacet ef to p. */
{
  double p1 = sos_fp (p, 1);
  double p2 = sos_fp (p, 2);
  double p3 = sos_fp (p, 3);
  double d1, d2, d3;
  int a, b, c;
  trist_triangle (ef, &a, &b, &c);
  d1 = (sos_fp (a, 1) + sos_fp (b, 1) + sos_fp (c, 1)) / 3.0;
  d2 = (sos_fp (a, 2) + sos_fp (b, 2) + sos_fp (c, 2)) / 3.0;
  d3 = (sos_fp (a, 3) + sos_fp (b, 3) + sos_fp (c, 3)) / 3.0;
  return (square (p1 - d1) + square (p2 - d2) + square (p3 - d3));
}

#endif

/*--------------------------------------------------------------------------*/

static double square (double x)
{
  return (x * x);
}    

/*--------------------------------------------------------------------------*/

int trist_search_get_tests (void)
     /* Returns (accumulated) number of sos_positive3() tests. */
{
  return (positive3_tests);
}

int trist_search_get_tets (void)
     /* Returns (accumulated) number of tetrahedra visited. */
{
  return (tets_visited);
}
